import logging
import random
import sys
import unittest
from time import time

from BTrees.IIBTree import weightedIntersection

from Products.PluginIndexes.BooleanIndex.BooleanIndex import BooleanIndex
from Products.PluginIndexes.CompositeIndex.CompositeIndex import CompositeIndex
from Products.PluginIndexes.FieldIndex.FieldIndex import FieldIndex
from Products.PluginIndexes.interfaces import ILimitedResultIndex
from Products.PluginIndexes.KeywordIndex.KeywordIndex import KeywordIndex


logger = logging.getLogger('zope.testCompositeIndex')

states = ['published', 'pending', 'private', 'intranet']
types = ['Document', 'News', 'File', 'Image']
default_pages = [True, False, False, False, False, False]
subjects = list(map(lambda x: 'subject_{0}'.format(x), range(6)))
keywords = list(map(lambda x: 'keyword_{0}'.format(x), range(6)))


class TestObject(object):

    def __init__(self, id, portal_type, review_state,
                 is_default_page=False, subject=(), keyword=()):
        self.id = id
        self.portal_type = portal_type
        self.review_state = review_state
        self.is_default_page = is_default_page
        self.subject = subject
        self.keyword = keyword

    def getPhysicalPath(self):
        return ['', self.id, ]

    def __repr__(self):
        return ('< {id}, {portal_type}, {review_state},\
        {is_default_page}, {subject} , {keyword}>'.format(
            id=self.id,
            portal_type=self.portal_type,
            review_state=self.review_state,
            is_default_page=self.is_default_page,
            subject=self.subject,
            keyword=self.keyword))


class RandomTestObject(TestObject):

    def __init__(self, id):

        i = random.randint(0, len(types) - 1)
        portal_type = types[i]

        i = random.randint(0, len(states) - 1)
        review_state = states[i]

        i = random.randint(0, len(default_pages) - 1)
        is_default_page = default_pages[i]

        subject = random.sample(subjects, random.randint(1, len(subjects)))
        keyword = random.sample(keywords, random.randint(1, len(keywords)))

        super(RandomTestObject, self).__init__(id, portal_type,
                                               review_state, is_default_page,
                                               subject, keyword)


# Pseudo ContentLayer class to support quick
# unit tests (skip performance tests)
class PseudoLayer(object):

    @classmethod
    def setUp(cls):
        pass

    @classmethod
    def tearDown(cls):
        pass


class CompositeIndexTestMixin(object):

    def setUp(self):
        self._indexes = [FieldIndex('review_state'),
                         FieldIndex('portal_type'),
                         BooleanIndex('is_default_page'),
                         KeywordIndex('subject',
                                      extra={
                                          'indexed_attrs':
                                          'keyword,subject'}
                                      ),
                         CompositeIndex('comp01',
                                        extra=[{'id': 'portal_type',
                                                'meta_type': 'FieldIndex',
                                                'attributes': ''},
                                               {'id': 'review_state',
                                                'meta_type': 'FieldIndex',
                                                'attributes': ''},
                                               {'id': 'is_default_page',
                                                'meta_type': 'BooleanIndex',
                                                'attributes': ''},
                                               {'id': 'subject',
                                                'meta_type': 'KeywordIndex',
                                                'attributes':
                                                'keyword,subject'}
                                               ])
                         ]

    def getIndex(self, name):
        for idx in self._indexes:
            if idx.id == name:
                return idx

    def defaultSearch(self, req, expectedValues=None, verbose=False):

        rs = None
        for index in self._indexes:
            st = time()
            duration = (time() - st) * 1000

            limit_result = ILimitedResultIndex.providedBy(index)
            if limit_result:
                r = index._apply_index(req, rs)
            else:
                r = index._apply_index(req)
            duration = (time() - st) * 1000

            if r is not None:
                r, u = r
                w, rs = weightedIntersection(rs, r)
                if not rs:
                    break

            if verbose and (index.id in req):
                logger.info('index %s: %s hits in %3.2fms',
                            index.id, r and len(r) or 0, duration)

        if not rs:
            return set()

        try:
            rs = rs.keys()
        except AttributeError:
            pass

        return set(rs)

    def compositeSearch(self, req, expectedValues=None, verbose=False):
        comp_index = self.getIndex('comp01')
        query = comp_index.make_query(req)

        # catch successful?
        self.assertIn('comp01', query)

        return self.defaultSearch(query,
                                  expectedValues=expectedValues,
                                  verbose=verbose)

    def enableLog(self):
        logger.root.setLevel(logging.INFO)
        logger.root.addHandler(logging.StreamHandler(sys.stdout))

    def populateIndexes(self, k, v):
        for index in self._indexes:
            index.index_object(k, v)

    def printIndexInfo(self):
        def info(index):
            size = index.indexSize()
            n_obj = index.numObjects()
            ratio = float(size) / float(n_obj)
            logger.info('<id: %15s unique keys: '
                        '%3s  length: %5s  ratio: %6.3f pm>',
                        index.id, size, n_obj, ratio * 1000)
            return ratio

        for index in self._indexes:
            info(index)

    def clearIndexes(self):
        for index in self._indexes:
            index.clear()


class CompositeIndexPerformanceTest(CompositeIndexTestMixin,
                                    unittest.TestCase):
    layer = PseudoLayer

    @unittest.skipIf(
        sys.platform.startswith('win'),
        'Time() is not well resolved in Windows.'
        ' In Python 3 use time.perf_count()')
    def testPerformance(self):
        self.enableLog()

        lengths = [10000, ]

        queries = [('query01_default_two_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'query': 'pending'}}),
                   ('query02_default_two_indexes',
                    {'portal_type': {'query': 'Document'},
                     'subject': {'query': 'subject_2'}}),
                   ('query02_default_two_indexes_zero_hits',
                    {'portal_type': {'query': 'Document'},
                     'subject': {'query': ['keyword_1', 'keyword_2']}}),
                   ('query03_default_two_indexes',
                    {'portal_type': {'query': 'Document'},
                     'subject': {'query': ['subject_1', 'subject_3']}}),
                   ('query04_default_two_indexes',
                    {'portal_type': {'query': 'Document'},
                     'is_default_page': {'query': False}}),
                   ('query05_default_two_indexes',
                    {'portal_type': {'query': 'Document'},
                     'is_default_page': {'query': True}}),
                   ('query06_default_two_indexes',
                    {'review_state': {'query': 'pending'},
                     'is_default_page': {'query': False}}),
                   ('query07_default_three_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': False}}),
                   ('query08_default_three_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': True}}),
                   ('query09_default_four_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': True},
                     'subject': {'query': ['subject_2', 'subject_3'],
                                 'operator': 'or'}}),
                   ('query10_and_operator_four_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': True},
                     'subject': {'query': ['subject_1', 'subject_3'],
                                 'operator': 'and'}}),
                   ('query11_and_operator_four_indexes',
                    {'portal_type': {'query': ('Document', 'News')},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': True},
                     'subject': {'query': ['subject_1', 'subject_3'],
                                 'operator': 'and'}}),
                   ('query12_not_operator_four_indexes',
                    {'portal_type': {'not': 'Document'},
                     'review_state': {'query': 'pending'},
                     'is_default_page': {'query': True},
                     'subject': {'query': ['subject_2', 'subject_3'],
                                 'operator': 'or'}}),
                   ('query13_not_operator_four_indexes',
                    {'portal_type': {'query': 'Document'},
                     'review_state': {'not': ('pending', 'visible')},
                     'is_default_page': {'query': True},
                     'subject': {'query': ['subject_2', 'subject_3']}}),
                   ]

        def profileSearch(query, warmup=False, verbose=False):

            st = time()
            res1 = self.defaultSearch(query, verbose=False)
            duration1 = (time() - st) * 1000

            if verbose:
                logger.info('atomic:    %s hits in %3.2fms',
                            len(res1), duration1)

            st = time()
            res2 = self.compositeSearch(query, verbose=False)
            duration2 = (time() - st) * 1000

            if verbose:
                logger.info('composite: %s hits in %3.2fms',
                            len(res2), duration2)

            if verbose:
                logger.info('[composite/atomic] factor %3.2f',
                            duration1 / duration2,)

            if not warmup:
                # if length of result is greater than zero composite
                # search must be roughly faster than default search
                if res1 and res2:
                    self.assertLess(
                        0.4 * duration2,
                        duration1,
                        (duration2, duration1, query))

            # is result identical?
            self.assertEqual(len(res1), len(res2), '{0} != {1} for {2}'.format(
                             len(res1), len(res2), query))
            self.assertEqual(res1, res2)

        for length in lengths:
            self.clearIndexes()
            logger.info('************************************\n'
                        'indexing %s objects', length)

            for i in range(length):
                name = str(i)
                obj = RandomTestObject(name)
                self.populateIndexes(i, obj)

            logger.info('indexing finished\n')

            self.printIndexInfo()

            logger.info('\nstart queries')

            # warming up indexes
            logger.info('warming up indexes')
            for name, query in queries:
                profileSearch(query, warmup=True)

            # in memory measure
            logger.info('in memory measure')
            for name, query in queries:
                logger.info('\nquery: %s', name)
                profileSearch(query, verbose=True)

            logger.info('\nqueries finished')

        logger.info('************************************')


class CompositeIndexTest(CompositeIndexTestMixin, unittest.TestCase):

    def testSearch(self):

        obj = TestObject('obj_1', 'Document', 'pending', subject=('subject_1'))
        self.populateIndexes(1, obj)
        obj = TestObject('obj_2', 'News', 'pending', subject=('subject_2'))
        self.populateIndexes(2, obj)
        obj = TestObject('obj_3', 'News', 'visible',
                         subject=('subject_1', 'subject_2'))
        self.populateIndexes(3, obj)
        obj = TestObject('obj_4', 'Event', 'private',
                         subject=('subject_1', 'subject_2'),
                         keyword=('keyword_1', ))
        self.populateIndexes(4, obj)

        queries = [
            # query on two attributes
            {'review_state': {'query': 'pending'},
             'portal_type': {'query': 'Document'}},
            # query on two attributes with 'or' operator
            {'review_state': {'query': ('pending', 'visible')},
             'portal_type': {'query': 'News'}},
            # query on two attributeswith one 'not' operator
            {'review_state': {'query': ('pending', 'visible')},
             'subject': {'query': 'subject_1', 'not': 'subject_2'}},
            # query on two attributes with 'or' operator
            {'review_state': {'query': 'pending'},
             'portal_type': {'query': ('News', 'Document')}},
            # query on three attributes
            {'review_state': {'query': ('pending', 'visible')},
             'portal_type': {'query': ('News', 'Document')},
             'is_default_page': {'query': False}},
            # query on four attributes with explicit 'or' operator
            {'review_state': {'query': ('pending', 'visible')},
             'portal_type': {'query': ('News', 'Document')},
             'is_default_page': {'query': False},
             'subject': {'query': ('subject_1', 'subject_2'),
                         'operator': 'or'}},
            # query on four attributes with explicit 'or' operator
            {'review_state': {'query': ('pending', 'visible')},
             'portal_type': {'query': ('News', 'Document')},
             'is_default_page': {'query': False},
             'subject': {'query': ('subject_1', 'subject_2'),
                         'operator': 'or'}},
            # query on four attributes with with explicit 'and' operator
            {'review_state': {'query': ('pending', 'visible')},
             'portal_type': {'query': ('News', 'Document')},
             'is_default_page': {'query': False},
             'subject': {'query': ('subject_1', 'subject_2'),
                         'operator': 'and'}},
            # query on five attributes with
            {'review_state': {'not': ('pending', 'visible')},
             'portal_type': {'query': ('News', 'Document')},
             'is_default_page': {'query': False},
             'subject': {'query': ('subject_1', )},
             'keyword': {'query': ('keyword_1',)}},
        ]

        for query in queries:
            res1 = self.defaultSearch(query)
            res2 = self.compositeSearch(query)
            # is result identical?
            self.assertEqual(len(res1), len(res2), '{0} != {1} for {2}'.format(
                len(res1), len(res2), query))

            self.assertEqual(res1, res2)

    def testMakeQuery(self):

        ci = CompositeIndex(
            'ci',
            extra=[
                dict(id='fi', meta_type='FieldIndex', attributes=('fi',)),
                dict(id='ki', meta_type='KeywordIndex', attributes=('ki',)),
                dict(id='bi', meta_type='BooleanIndex', attributes=('bi',)),
            ]
        )

        # avoid premature return of `make_query`
        ci._length.change(1)

        #
        # 'range' parameter not supported
        query = dict(fi=dict(query=(1, 3), range='min:max'),
                     ki=dict(query=(10, 11)))

        result = ci.make_query(query)

        # should return original query
        self.assertEqual(result, query)
        #

        # 'and' operator not supported
        query = dict(fi=dict(query=(1, 3)),
                     ki=dict(query=(10, 11), operator='and'))

        result = ci.make_query(query)

        # should return original query
        self.assertEqual(result, query)

        #
        # regular query
        query = dict(fi=dict(query=(1, 2, 3)),
                     ki=dict(query=(10, 11)))

        result = ci.make_query(query)

        expect = {'ci': {'query': ((('fi', 1), ('ki', 10)),
                                   (('fi', 1), ('ki', 11)),
                                   (('fi', 2), ('ki', 10)),
                                   (('fi', 2), ('ki', 11)),
                                   (('fi', 3), ('ki', 10)),
                                   (('fi', 3), ('ki', 11)))}}

        self.assertEqual(set(result['ci']['query']),
                         set(expect['ci']['query']))

        #
        # 'not' parameter in query
        query = dict(fi=dict(query=(1, 2, 3)),
                     ki={'query': (10, 11), 'not': 15})

        result = ci.make_query(query)

        expect = {'ci': {'not': ((('fi', 1), ('ki', 15)),
                                 (('fi', 2), ('ki', 15)),
                                 (('fi', 3), ('ki', 15))),
                         'query': ((('fi', 1), ('ki', 10)),
                                   (('fi', 1), ('ki', 11)),
                                   (('fi', 2), ('ki', 10)),
                                   (('fi', 2), ('ki', 11)),
                                   (('fi', 3), ('ki', 10)),
                                   (('fi', 3), ('ki', 11)))}}

        # sequence of result is not deterministic.
        # Therefore, we use type 'set' for comparison.
        self.assertEqual(set(result['ci']['not']),
                         set(expect['ci']['not']))
        self.assertEqual(set(result['ci']['query']),
                         set(expect['ci']['query']))

        #
        # 'pure not' is not supported. The affected attribute cannot
        # be used to optimize the query.

        # In the case of two attributes the query cannot be optimized.
        query = dict(fi=dict(query=(1, 2, 3)),
                     ki={'not': (15, 16)})

        result = ci.make_query(query)

        # should return original query
        self.assertEqual(result, query)

        # In the case of three attributes only two attributes can
        # be used to optimize the query.
        query = dict(fi={'query': (1, 2, 3)},
                     ki={'not': (15, 16)},
                     bi={'query': (True,)})

        result = ci.make_query(query)

        expect = {'ci': {'query': ((('fi', 1), ('bi', 1)),
                                   (('fi', 2), ('bi', 1)),
                                   (('fi', 3), ('bi', 1)))},
                  'ki': {'not': (15, 16)}}

        # sequence of result is not deterministic.
        # Therefore, we use type 'set' for comparison.
        self.assertEqual(set(result['ki']['not']),
                         set(expect['ki']['not']))
        self.assertEqual(set(result['ci']['query']),
                         set(expect['ci']['query']))

        #
        # 'pure not' query on any attribute is also not supported
        query = dict(fi={'not': 1},
                     ki={'not': (15, 16)})

        result = ci.make_query(query)

        # should return original query
        self.assertEqual(result, query)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import print_function

import sys
import time
from logging import getLogger

import transaction
from DateTime.DateTime import DateTime
from zope.interface import implementer

from .interfaces import IProgressHandler


LOG = getLogger('ProgressHandler')


@implementer(IProgressHandler)
class StdoutHandler(object):
    """ A simple progress handler """

    def __init__(self, steps=100):
        self._steps = steps

    def init(self, ident, max, savepoint=True):
        self._ident = ident
        self._max = max
        self.savepoint = savepoint
        self._start = time.time()
        self.fp = sys.stdout
        self.output('Process started (%d objects to go)' % self._max)

    def info(self, text):
        self.output(text)

    def finish(self):
        self.output('Process terminated. Duration: %0.2f seconds' %
                    (time.time() - self._start))

    def report(self, current, *args, **kw):
        if current > 0:
            if current % self._steps == 0:
                if self.savepoint:
                    transaction.savepoint(optimistic=True)
                seconds_so_far = time.time() - self._start
                seconds_to_go = (seconds_so_far
                                 / current
                                 * (self._max - current))
                end = DateTime(time.time() + seconds_to_go)
                self.output('%d/%d (%.2f%%) Estimated termination: %s' %
                            (current, self._max, (100.0 * current / self._max),
                             end.strftime('%Y/%m/%d %H:%M:%Sh')))

    def output(self, text):
        print('%s: %s' % (self._ident, text), file=self.fp)


class ZLogHandler(StdoutHandler):
    """ Use Zope logger"""

    def output(self, text):
        LOG.info(text)


class FilelogHandler(StdoutHandler):
    """ Use a custom file for logging """

    def __init__(self, filename, steps=100):
        StdoutHandler.__init__(self, steps)
        self.filename = filename

    def output(self, text):
        with open(self.filename, 'a') as fd:
            fd.write(text + '\n')



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from ZPublisher.HTTPRequest import record


_marker = object()


class IndexQueryParseError(Exception):
    pass


class IndexQuery(object):
    """
    This class provides functionality to hide the internals of a query
    send from the Catalog/ZCatalog to an index._apply_index() method.

    The class understands the following type of parameters:

    - old-style parameters where the query for an index as value inside
      the request dictionary where the index name is the name of the key.

    - dictionary-style parameters specify a query for an index as
      an entry in the request dictionary where the key corresponds to the
      name of the index and the key is a dictionary with the parameters
      passed to the index.

      Allowed keys of the parameter dictionary:

      'query'  - contains the query (either string, list or tuple) (required)

      other parameters depend on the the index.
    """
    operators = ('or', 'and')

    def __init__(self, request, iid, options=(), operators=('or', 'and'),
                 default_operator='or'):
        """Parse a query from the ZPublisher and return a uniform
        datastructure back to the _apply_index() method of the index.

          query -- the query dictionary send from the ZPublisher
          iid     -- Id of index
          options -- a list of options the index is interested in
          operators -- a tuple of allowed operators
          default_operator -- the default operator
        """

        self.id = iid
        self.operators = operators
        self.operator = default_operator
        self.options = options

        if iid not in request:
            self.keys = None
            return

        param = request[iid]
        keys = None

        if isinstance(param, (dict, record)):
            # query is a dictionary containing all parameters
            query = param.get('query', ())
            if isinstance(query, (tuple, list)):
                keys = query
            else:
                keys = [query]

            for op in param.keys():
                if op == 'query':
                    continue

                self.set(op, param[op])

        else:
            # query is tuple, list, string, number, or something else
            if isinstance(param, (tuple, list)):
                keys = param
            else:
                keys = [param]

            for field in request.keys():
                if field.startswith(iid + '_'):
                    op = field[len(iid) + 1:]
                    self.set(op, request[field])

        self.keys = keys
        not_value = getattr(self, 'not', None)
        if not_value is not None:
            if not isinstance(not_value, (tuple, list)):
                not_value = [not_value]
                self.set('not', not_value)

    @property
    def operator(self):
        return self._operator

    @operator.setter
    def operator(self, value):
        iid = self.id
        value = value.lower()
        if value not in self.operators:
            raise ValueError(('index {0!r}: operator {1!r}'
                              ' is not valid').format(iid, value))
        self._operator = value

    def get(self, key, default_v=None):
        value = getattr(self, key, _marker)
        if value is not _marker:
            return value
        return default_v

    def set(self, key, value):
        if key in self.options:
            setattr(self, key, value)
        else:
            raise ValueError(('index {0!r}: option {1!r}'
                              ' is not valid').format(self.id, key))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from Acquisition import Implicit
from Acquisition import aq_base
from Acquisition import aq_parent
from zExceptions import Unauthorized
from ZODB.POSException import ConflictError
from zope.globalrequest import clearRequest
from zope.globalrequest import setRequest


_marker = object()


class Happy(Implicit):
    """Happy content"""

    def __init__(self, id):
        self.id = id

    def check(self):
        pass


class Secret(Happy):
    """Object that raises Unauthorized when accessed"""

    def check(self):
        raise Unauthorized


class Conflicter(Happy):
    """Object that raises ConflictError when accessed"""

    def check(self):
        raise ConflictError


class DummyRequest(object):

    def physicalPathToURL(self, path, relative=False):
        if not relative:
            path = 'http://superbad.com' + path
        return path


class DummyRoot(Implicit):

    happy = Happy('happy')
    secret = Secret('secret')
    conflicter = Conflicter('conflicter')

    def unrestrictedTraverse(self, path, default=None):
        assert path in ['', ('',), ['']], path
        return self

    def restrictedTraverse(self, path, default=_marker):
        if path.startswith('/'):
            path = path[1:]
        try:
            ob = getattr(self, path).__of__(self)
            ob.check()
            return ob
        except AttributeError:
            if default is not _marker:
                return default
            raise


class DummyCatalog(Implicit):

    _paths = ['/conflicter', '/happy', '/secret', '/zonked']

    def getPhysicalRoot(self):
        return aq_parent(self)

    def getpath(self, rid):
        return self._paths[rid]

    def getobject(self, rid):
        return self.restrictedTraverse(self._paths[rid])

    def resolve_url(self, path, REQUEST):
        # strip server part
        path = path[path.find('/', path.find('//') + 1):]
        return self.restrictedTraverse(path)


class ConflictingCatalog(DummyCatalog):

    def getpath(self, rid):
        raise ConflictError


class TestBrains(unittest.TestCase):

    def setUp(self):
        self.root = DummyRoot()
        self.cat = DummyCatalog().__of__(self.root)

    def _makeBrain(self, rid):
        from Products.ZCatalog.CatalogBrains import AbstractCatalogBrain

        class Brain(AbstractCatalogBrain):
            __record_schema__ = {'test_field': 0, 'data_record_id_': 1}

        return Brain(('test', rid)).__of__(self.cat)

    def testHasKey(self):
        b = self._makeBrain(1)
        self.assertIn('test_field', b)
        self.assertIn('data_record_id_', b)
        self.assertNotIn('godel', b)

    def testGetPath(self):
        b = [self._makeBrain(rid) for rid in range(3)]
        self.assertEqual(b[0].getPath(), '/conflicter')
        self.assertEqual(b[1].getPath(), '/happy')
        self.assertEqual(b[2].getPath(), '/secret')

    def testGetPathPropagatesConflictErrors(self):
        self.cat = ConflictingCatalog()
        b = self._makeBrain(0)
        self.assertRaises(ConflictError, b.getPath)

    def testGetURL(self):
        request = DummyRequest()
        b = self._makeBrain(0)

        self.root.REQUEST = request
        self.assertEqual(b.getURL(), 'http://superbad.com/conflicter')

    def testGetURL_catalog_as_utility(self):
        request = DummyRequest()
        b = self._makeBrain(0)

        setRequest(request)
        self.assertEqual(b.getURL(), 'http://superbad.com/conflicter')
        clearRequest()

    def testGetRID(self):
        b = self._makeBrain(42)
        self.assertEqual(b.getRID(), 42)

    def testGetObjectHappy(self):
        request = DummyRequest()
        b = self._makeBrain(1)

        self.root.REQUEST = request
        self.assertEqual(b.getPath(), '/happy')
        self.assertEqual(b.getObject().REQUEST, request)
        self.assertTrue(aq_base(b.getObject()) is
                        aq_base(self.cat.getobject(1)))

    def testGetObjectHappy_catalog_as_utility(self):
        request = DummyRequest()
        b = self._makeBrain(1)

        setRequest(request)
        self.assertEqual(b.getPath(), '/happy')
        self.assertEqual(b.getObject().REQUEST, request)
        self.assertTrue(aq_base(b.getObject()) is
                        aq_base(self.cat.getobject(1)))
        clearRequest()

    def testGetObjectPropagatesConflictErrors(self):
        b = self._makeBrain(0)
        self.assertEqual(b.getPath(), '/conflicter')
        self.assertRaises(ConflictError, b.getObject)

    def testGetObjectRaisesUnauthorized(self):
        b = self._makeBrain(2)
        self.assertEqual(b.getPath(), '/secret')
        self.assertRaises(Unauthorized, b.getObject)

    def testGetObjectRaisesNotFoundForMissing(self):
        from zExceptions import NotFound
        b = self._makeBrain(3)
        self.assertEqual(b.getPath(), '/zonked')
        self.assertRaises(AttributeError, self.cat.getobject, 3)
        self.assertRaises((NotFound, AttributeError, KeyError), b.getObject)



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import logging
import time
from itertools import combinations
from itertools import product

from six.moves import urllib

import transaction
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.special_dtml import DTMLFile
from BTrees.OOBTree import OOSet
from BTrees.OOBTree import difference
from Persistence import PersistentMapping
from zope.interface import implementer

from Products.PluginIndexes.interfaces import ITransposeQuery
from Products.PluginIndexes.KeywordIndex.KeywordIndex import KeywordIndex
from Products.PluginIndexes.unindex import _marker
from Products.ZCatalog.query import IndexQuery


try:
    from time import clock as process_time
except ImportError:
    from time import process_time


LOG = logging.getLogger('CompositeIndex')

QUERY_OPTIONS = {
    'BooleanIndex': ('query', 'range', 'not'),
    'FieldIndex': ('query', 'range', 'not'),
    'KeywordIndex': ('query', 'range', 'not', 'operator'),
}
QUERY_OPERATORS = {
    'BooleanIndex': (('and', 'or'), 'or'),
    'FieldIndex': (('and', 'or'), 'or'),
    'KeywordIndex': (('and', 'or'), 'or'),
}

MIN_COMPONENTS = 2


def tuple_cast(kw):
    if isinstance(kw, list):
        kw = tuple(kw)
    elif not isinstance(kw, tuple):
        kw = (kw,)
    return kw


def collect(recs, not_cid=None):
    kw_list = []
    for c_id, rec in recs:
        if c_id != not_cid:
            keys = list(rec.keys)
            if not_cid is not None:
                keys.extend(rec.get('not', []))
        elif c_id == not_cid:
            keys = rec.get('not', [])
        if not keys:
            continue
        keys = tuple_cast(keys)
        kw = tuple([(c_id, k) for k in keys])
        kw_list.append(kw)
    return kw_list


class ComponentMapping(PersistentMapping):
    """A persistent wrapper for mapping objects
    recording the order in which items are added. """

    def __init__(self, *args, **kwargs):
        self._keys = []
        PersistentMapping.__init__(self, *args, **kwargs)

    def __delitem__(self, key):
        self._keys.remove(key)
        PersistentMapping.__delitem__(self, key)

    def __setitem__(self, key, item):
        if key not in self._keys:
            self._keys.append(key)
        PersistentMapping.__setitem__(self, key, item)

    def clear(self):
        self._keys = []
        PersistentMapping.clear(self)

    def copy(self):
        cm = ComponentMapping()
        cm.update(self)
        return cm

    def items(self):
        return zip(self._keys, self.values())

    def keys(self):
        return self._keys[:]

    def popitem(self):
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError('dictionary is empty')

        val = self[key]
        del self[key]

        return (key, val)

    def setdefault(self, key, failobj=None):
        if key not in self._keys:
            self._keys.append(key)
        return PersistentMapping.setdefault(self, key, failobj)

    def update(self, d):
        for (key, val) in d.items():
            self.__setitem__(key, val)

    def values(self):
        return list(map(self.get, self._keys))


class Component(object):

    _attributes = ''

    def __init__(self, id, meta_type, attributes):
        self._id = id
        self._meta_type = meta_type
        if attributes:
            self._attributes = attributes

    @property
    def id(self):
        return self._id

    @property
    def meta_type(self):
        return self._meta_type

    @property
    def attributes(self):

        attributes = self._attributes

        if not attributes:
            return [self._id]

        if isinstance(attributes, str):
            return attributes.split(',')

        attributes = list(attributes)

        attributes = [attr.strip() for attr in attributes if attr]

        return attributes

    @property
    def rawAttributes(self):
        return self._attributes

    def __repr__(self):
        return ('<id: {0.id}; metatype: {0.meta_type}; '
                'attributes: {0.attributes}>').format(self)


@implementer(ITransposeQuery)
class CompositeIndex(KeywordIndex):

    """Index for composition of simple fields.
       or sequences of items
    """

    meta_type = 'CompositeIndex'

    manage_options = (
        {'label': 'Settings',
         'action': 'manage_main'},
        {'label': 'Browse',
         'action': 'manage_browse'},
    )

    query_options = ('query', 'operator', 'not')

    def __init__(self, id, ignore_ex=None, call_methods=None,
                 extra=None, caller=None):
        """Create an composite index"""

        self.id = id
        self.ignore_ex = ignore_ex        # currently unimplemented
        self.call_methods = call_methods

        # set components
        self._components = ComponentMapping()
        if extra:
            for cdata in extra:
                c_id = cdata['id']
                c_meta_type = cdata['meta_type']
                c_attributes = cdata['attributes']
                self._components[c_id] = Component(c_id, c_meta_type,
                                                   c_attributes)
        self.clear()

    def _index_object(self, documentId, obj, threshold=None, attr=''):

        # get permuted keywords
        newKeywords = self._get_permuted_keywords(obj)

        oldKeywords = self._unindex.get(documentId, None)

        if oldKeywords is None:
            # we've got a new document, let's not futz around.
            try:
                for kw in newKeywords:
                    self.insertForwardIndexEntry(kw, documentId)
                if newKeywords:
                    self._unindex[documentId] = list(newKeywords)
            except TypeError:
                return 0
        else:
            # we have an existing entry for this document, and we need
            # to figure out if any of the keywords have actually changed
            if type(oldKeywords) is not OOSet:
                oldKeywords = OOSet(oldKeywords)
            newKeywords = OOSet(newKeywords)
            fdiff = difference(oldKeywords, newKeywords)
            rdiff = difference(newKeywords, oldKeywords)
            if fdiff or rdiff:
                # if we've got forward or reverse changes
                if newKeywords:
                    self._unindex[documentId] = list(newKeywords)
                else:
                    del self._unindex[documentId]
                if fdiff:
                    self.unindex_objectKeywords(documentId, fdiff)
                if rdiff:
                    for kw in rdiff:
                        self.insertForwardIndexEntry(kw, documentId)
        return 1

    def _get_permuted_keywords(self, obj):
        """ returns permutation tuple of object keywords """

        components = self.getIndexComponents()
        kw_list = []

        for c in components:
            kw = self._get_component_keywords(obj, c)
            # skip if keyword list is empty
            if not kw:
                continue
            kw = tuple([(c.id, k) for k in kw])
            kw_list.append(kw)

        # permute keyword list in order to support any combination and
        # number (n > 1) of components in query
        pkl = []
        c_list = product(*kw_list)

        for c in c_list:
            for r in range(MIN_COMPONENTS, len(c) + 1):
                p = combinations(c, r)
                pkl.extend(p)

        return tuple(pkl)

    def _get_component_keywords(self, obj, component):

        if component.meta_type == 'FieldIndex':
            # last attribute is the winner if value is not None
            for attr in component.attributes:
                datum = self._get_object_datum(obj, attr)
                if datum is None:
                    continue
            if datum is None:
                return ()
            if isinstance(datum, list):
                datum = tuple(datum)
            return (datum,)

        elif component.meta_type == 'KeywordIndex':
            # last attribute is the winner
            attr = component.attributes[-1]
            datum = self._get_object_keywords(obj, attr)
            if isinstance(datum, list):
                datum = tuple(datum)
            return datum

        elif component.meta_type == 'BooleanIndex':
            # last attribute is the winner
            attr = component.attributes[-1]
            datum = self._get_object_datum(obj, attr)
            if datum is not _marker:
                datum = int(bool(datum))
            return (datum,)

        else:
            raise KeyError

    def getIndexComponents(self):
        """ return sequence of indexed attributes """
        return self._components.values()

    def getComponentIndexNames(self):
        """ returns component index names to composite """

        return tuple([c.id for c in self.getIndexComponents()])

    def getComponentIndexAttributes(self):
        """ returns list of attributes of each component index to composite"""

        return tuple([c.attributes for c in self.getIndexComponents()])

    def getIndexNames(self):
        """ returns index names that are caught by query substitution """
        return self.getComponentIndexNames()

    def make_query(self, query):
        """ optimize the query for supported index names """

        try:
            zc = aq_parent(aq_parent(self))
            skip = zc.getProperty('skip_compositeindex', False)
            if skip:
                LOG.debug('%(context)s: skip composite query build '
                          'for %(zcatalog)r', dict(
                              context=self.__class__.__name__,
                              zcatalog=zc))
                return query
        except AttributeError:
            pass

        if len(self) == 0:
            return query

        cquery = query.copy()
        components = self.getIndexComponents()

        # collect components matching query attributes
        # and check them for completeness
        c_records = []

        # component ids containing 'not' operator
        not_cids = []

        for c in components:
            query_options = QUERY_OPTIONS[c.meta_type]
            query_operators = QUERY_OPERATORS[c.meta_type]
            rec = IndexQuery(query, c.id, query_options,
                             query_operators[0], query_operators[1])
            opr = None

            # not supported: 'range' parameter
            range_parm = rec.get('range', None)
            if range_parm:
                opr = 'range'
            if rec.get('usage', None):
                # see if any usage params are sent to field
                opr = rec.usage.lower().split(':')
            if opr == 'range':
                continue

            # not supported: 'and' operator
            if rec.keys and rec.operator == 'and':
                continue

            # continue if no keys in query were set
            if rec.keys is None:
                continue

            # convert rec keys to int for BooleanIndex
            if c.meta_type == 'BooleanIndex':
                rec.keys = [int(bool(v)) for v in rec.keys[:]]

            # rec with 'not' parameter
            not_parm = rec.get('not', None)
            if not_parm:
                # not supported: 'pure not'
                if len(rec.keys) == 0:
                    continue
                not_cids.append(c.id)
                if c.meta_type == 'BooleanIndex':
                    not_parm = [int(bool(v)) for v in not_parm[:]]
                    rec.set('not', not_parm)

            c_records.append((c.id, rec))

        # return if less than MIN_COMPONENTS query attributes were caught
        if len(c_records) < MIN_COMPONENTS:
            return query

        records = ()
        kw_list = collect(c_records)
        # permute keyword list
        records = tuple(product(*kw_list))

        not_records = set()
        for c_id in not_cids:
            kw_list = collect(c_records, not_cid=c_id)
            if kw_list:
                not_records.update(product(*kw_list))
        # permute keyword list for 'not' operator
        not_records = tuple(not_records)

        # substitute matching query attributes as composite index
        if records and not_records:
            cquery.update({self.id: {'query': records, 'not': not_records}})
        elif records:
            cquery.update({self.id: {'query': records}})
        elif not_records:
            cquery.update({self.id: {'not': not_records}})

        # delete original matching query attributes from query
        for c_id, rec in c_records:
            if c_id in cquery:
                del cquery[c_id]

        return cquery

    def addComponent(self, c_id, c_meta_type, c_attributes):
        # Add a component object by 'c_id'.
        if c_id in self._components:
            raise KeyError('A component with this '
                           'name already exists: {0}'.format(c_id))

        self._components[c_id] = Component(c_id,
                                           c_meta_type,
                                           c_attributes)
        self.clear()

    def delComponent(self, c_id):
        # Delete the component object specified by 'c_id'.
        if c_id not in self._components:
            raise KeyError('no such Component:  {0}'.format(c_id))

        del self._components[c_id]

        self.clear()

    def saveComponents(self, components):
        # Change the component object specified by 'c_id'.
        for c in components:
            self.delComponent(c.old_id)
            self.addComponent(c.id, c.meta_type, c.attributes)

        # better safe than sorry
        self.clear()

    def manage_addComponent(self, c_id, c_meta_type, c_attributes, URL1,
                            REQUEST=None, RESPONSE=None):
        """ add a new component """
        if len(c_id) == 0:
            raise RuntimeError('Length of component ID too short')
        if len(c_meta_type) == 0:
            raise RuntimeError('No component type set')

        self.addComponent(c_id, c_meta_type, c_attributes)

        if RESPONSE:
            RESPONSE.redirect(URL1 + '/manage_main?'
                              'manage_tabs_message=Component%20added')

    def manage_delComponents(self, del_ids=(), URL1=None,
                             REQUEST=None, RESPONSE=None):
        """ delete one or more components """
        if not del_ids:
            raise RuntimeError('No component selected')

        for c_id in del_ids:
            self.delComponent(c_id)

        if RESPONSE:
            RESPONSE.redirect(URL1 + '/manage_main?'
                              'manage_tabs_message=Component(s)%20deleted')

    def manage_saveComponents(self, components, URL1=None,
                              REQUEST=None, RESPONSE=None):
        """ save values of components """

        self.saveComponents(components)

        if RESPONSE:
            RESPONSE.redirect(URL1 + '/manage_main?'
                              'manage_tabs_message=Component(s)%20updated')

    def fastBuild(self, threshold=None):

        if threshold is None:
            threshold = 10000

        zc = aq_parent(aq_parent(aq_inner(self)))
        getIndex = zc._catalog.getIndex
        components = self.getIndexComponents()

        self.clear()

        class pseudoObject(object):
            pass

        counter = 0
        for rid in zc._catalog.paths.keys():
            # pseudo object
            obj = pseudoObject()
            for c in components:
                kw = getIndex(c.id).getEntryForObject(rid, _marker)
                if kw is not _marker:
                    for attr in c.attributes:
                        setattr(obj, attr, kw)

            self.index_object(rid, obj)
            del obj

            counter += 1
            if counter > threshold:
                transaction.savepoint(optimistic=True)
                self._p_jar.cacheGC()
                counter = 0

    def manage_fastBuild(self, threshold=None, URL1=None,
                         REQUEST=None, RESPONSE=None):
        """ fast build index directly via catalog brains and attribute values
            of matching field and keyword indexes """

        tt = time.time()
        ct = process_time()

        self.fastBuild(threshold)

        tt = time.time() - tt
        ct = process_time() - ct

        if RESPONSE:
            msg = ('ComponentIndex fast reindexed '
                   'in {0:.3f}s ({1:.3f}s cpu time)').format(tt, ct)
            param = urllib.parse.urlencode({'manage_tabs_message': msg})

            RESPONSE.redirect(URL1 + '/manage_main?' + param)

    manage = manage_main = DTMLFile('dtml/manageCompositeIndex', globals())
    manage_main._setName('manage_main')
    manage_browse = DTMLFile('../dtml/browseIndex', globals())


manage_addCompositeIndexForm = DTMLFile('dtml/addCompositeIndex', globals())


def manage_addCompositeIndex(self, id, extra=None, REQUEST=None,
                             RESPONSE=None, URL3=None):
    """Add a composite index"""
    return self.manage_addIndex(id, 'CompositeIndex', extra=extra,
                                REQUEST=REQUEST, RESPONSE=RESPONSE, URL1=URL3)



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PluginIndexes interfaces.
"""

from zope.interface import Attribute
from zope.interface import Interface
from zope.schema import Bool


class IPluggableIndex(Interface):

    def getId():
        """Return Id of index."""

    def getEntryForObject(documentId, default=None):
        """Get all information contained for 'documentId'."""

    def getIndexSourceNames():
        """Get a sequence of attribute names that are indexed by the index.
        """

    def getIndexQueryNames():
        """Get a sequence of query parameter names to which this index applies.
        """

    def index_object(documentId, obj, threshold=None):
        """Index an object.

        - ``documentId`` is the integer ID of the document.

        - ``obj`` is the object to be indexed.

        - ``threshold`` is the number of words to process between committing
          subtransactions.  If None, subtransactions are disabled.

        For each name in ``getIndexSourceNames``, try to get the named
        attribute from ``obj``.

        - If the object does not have the attribute, do not add it to the
          index for that name.

        - If the attribute is a callable, call it to get the value.  If
          calling it raises an AttributeError, do not add it to the index.
          for that name.
        """

    def unindex_object(documentId):
        """Remove the documentId from the index."""

    def _apply_index(request):
        """Apply the index to query parameters given in 'request'.

        The argument should be a mapping object.

        If the request does not contain the needed parameters, then
        None is returned.

        If the request contains a parameter with the name of the
        column and this parameter is either a Record or a class
        instance then it is assumed that the parameters of this index
        are passed as attribute (Note: this is the recommended way to
        pass parameters since Zope 2.4)

        Otherwise two objects are returned.  The first object is a
        ResultSet containing the record numbers of the matching
        records.  The second object is a tuple containing the names of
        all data fields used.
        """

    def numObjects():
        """Return the number of indexed objects."""

    def indexSize():
        """Return the size of the index in terms of distinct values."""

    def clear():
        """Empty the index"""


class ILimitedResultIndex(IPluggableIndex):

    def _apply_index(request, resultset=None):
        """Same as IPluggableIndex' _apply_index method.

        If not `None`, the additional resultset argument
        indicates that the search result is relevant only on this set,
        i.e. everything outside resultset is of no importance.
        The index can use this information for optimizations.
        """


class IQueryIndex(IPluggableIndex):

    id = Attribute('Index id used to query the index.')
    operators = Attribute('A tuple of known/allowed operators.')
    useOperator = Attribute('A string specifying the default operator.')
    query_options = Attribute('Supported query options for the index.')

    def query_index(record, resultset=None):
        """Same as _apply_index, but the query is already a pre-parsed
        IndexQuery object.
        """


class IUniqueValueIndex(IPluggableIndex):
    """An index which can return lists of unique values contained in it"""

    def hasUniqueValuesFor(name):
        """Return true if the index can return the unique values for name"""

    def uniqueValues(name=None, withLengths=0):
        """Return an iterable/sequence of unique values for name.

        If 'withLengths' is true, returns a iterable/sequence of tuples of
        (value, length).
        """


class ISortIndex(IPluggableIndex):
    """An index which may be used to sort a set of document ids"""

    def keyForDocument(documentId):
        """Return the sort key that cooresponds to the specified document id

        This method is no longer used by ZCatalog, but is left for backwards
        compatibility."""

    def documentToKeyMap():
        """Return an object that supports __getitem__ and may be used to
        quickly lookup the sort key given a document id"""


class IDateIndex(Interface):

    """Index for dates.
    """

    index_naive_time_as_local = Bool(title=u'Index naive time as local?')


class IDateRangeIndex(Interface):

    """Index for date ranges, such as the "effective-expiration" range in CMF.

    Any object may return None for either the start or the end date: for the
    start date, this should be the logical equivalent of "since the beginning
    of time"; for the end date, "until the end of time".

    Therefore, divide the space of indexed objects into four containers:

    - Objects which always match (i.e., they returned None for both);

    - Objects which match after a given time (i.e., they returned None for the
      end date);

    - Objects which match until a given time (i.e., they returned None for the
      start date);

    - Objects which match only during a specific interval.
    """

    def getSinceField():
        """Get the name of the attribute indexed as start date.
        """

    def getUntilField():
        """Get the name of the attribute indexed as end date.
        """


class IPathIndex(Interface):

    """Index for paths returned by getPhysicalPath.

    A path index stores all path components of the physical path of an object.

    Internal datastructure:

    - a physical path of an object is split into its components

    - every component is kept as a  key of a OOBTree in self._indexes

    - the value is a mapping 'level of the path component' to
      'all docids with this path component on this level'
    """
    def insertEntry(comp, id, level):
        """ Insert an entry.

        This method is intended for use by subclasses:  it is not
        a normal API for the index.

        'comp' is an individual path component

        'id' is the docid

        .level'is the level of the component inside the path
        """


class IFilteredSet(Interface):

    """A pre-calculated result list based on an expression.
    """

    def getExpression():
        """Get the expression.
        """

    def getIds():
        """Get the IDs of all objects for which the expression is True.
        """

    def setExpression(expr):
        """Set the expression.
        """


class ITopicIndex(Interface):

    """A TopicIndex maintains a set of FilteredSet objects.

    Every FilteredSet object consists of an expression and and IISet with all
    Ids of indexed objects that eval with this expression to 1.
    """

    def addFilteredSet(filter_id, typeFilteredSet, expr):
        """Add a FilteredSet object.
        """

    def delFilteredSet(filter_id):
        """Delete the FilteredSet object specified by 'filter_id'.
        """

    def clearFilteredSet(filter_id):
        """Clear the FilteredSet object specified by 'filter_id'.
        """

# IIndexConfiguration was added on request by the GenericSetup community in
# order to perform introspection on indexes in a defined way.
# (ajung)


class IIndexConfiguration(Interface):
    """ Introspection API for pluggable indexes """

    def getSettings(self):
        """ Returns an mapping with index specific settings.
            E.g. {'indexed_attrs' : ('SearchableText', )}.
            The interface does not define any specifc mapping keys.
        """


class IRequestCacheIndex(Interface):
    """ Request cache API for pluggable indexes """

    def getRequestCache():
        """ Returns dict for caching per request for interim results
            of an index search. Returns 'None' if no REQUEST attribute
            is available
        """

    def getRequestCacheKey(record, resultset=None):
        """ Returns an unique key of a search record """


# ITransposeQuery was added to support multicolumn indexes for field and
# keyword indexes, eg. CompositeIndex. (andbag)
class ITransposeQuery(Interface):
    """ Optimization API for queries and indexing """

    def make_query(query):
        """ returns an optimized query for given index """

    def getIndexNames():
        """ returns index names that are optimized by index """



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from logging import getLogger

from Acquisition import aq_base
from App.special_dtml import DTMLFile
from BTrees.IOBTree import IOBTree
from BTrees.Length import Length
from BTrees.OIBTree import OIBTree

from Products.PluginIndexes.unindex import UnIndex
from Products.PluginIndexes.unindex import _marker


logger = getLogger('Products.ZCatalog')


class UUIDIndex(UnIndex):
    """Index for uuid fields with an unique value per key.

    The internal structure is:

    self._index = {datum:documentId]}
    self._unindex = {documentId:datum}

    For each datum only one documentId can exist.
    """

    meta_type = 'UUIDIndex'

    manage_options = (
        {'label': 'Settings', 'action': 'manage_main'},
        {'label': 'Browse', 'action': 'manage_browse'},
    )

    query_options = ['query', 'range', 'not']

    manage = manage_main = DTMLFile('dtml/manageUUIDIndex', globals())
    manage_main._setName('manage_main')
    manage_browse = DTMLFile('../dtml/browseIndex', globals())

    def clear(self):
        self._length = Length()
        self._index = OIBTree()
        self._unindex = IOBTree()
        if self._counter is None:
            self._counter = Length()
        else:
            self._increment_counter()

    def numObjects(self):
        """Return the number of indexed objects. Since we have a 1:1 mapping
        from documents to values, we can reuse the stored length.
        """
        return self.indexSize()

    def uniqueValues(self, name=None, withLengths=0):
        """returns the unique values for name

        if withLengths is true, returns a sequence of
        tuples of (value, length)
        """
        if name is None:
            name = self.id
        elif name != self.id:
            return

        if not withLengths:
            for key in self._index.keys():
                yield key
        else:
            # We know the length for each value is one
            for key in self._index.keys():
                yield (key, 1)

    def insertForwardIndexEntry(self, entry, documentId):
        """Take the entry provided and put it in the correct place
        in the forward index.
        """
        if entry is None:
            return

        old_docid = self._index.get(entry, _marker)
        if old_docid is _marker:
            self._index[entry] = documentId
            self._length.change(1)
        elif old_docid != documentId:
            logger.error("A different document with value '%s' already "
                         'exists in the index.', entry)

    def removeForwardIndexEntry(self, entry, documentId):
        """Take the entry provided and remove any reference to documentId
        in its entry in the index.
        """
        old_docid = self._index.get(entry, _marker)
        if old_docid is not _marker:
            del self._index[entry]
            self._length.change(-1)

    def _get_object_datum(self, obj, attr):
        # for a uuid it never makes sense to acquire a parent value via
        # Acquisition
        has_attr = getattr(aq_base(obj), attr, _marker)
        if has_attr is _marker:
            return _marker
        return super(UUIDIndex, self)._get_object_datum(obj, attr)


manage_addUUIDIndexForm = DTMLFile('dtml/addUUIDIndex', globals())


def manage_addUUIDIndex(self, id, extra=None,
                        REQUEST=None, RESPONSE=None, URL3=None):
    """Add an uuid index"""
    return self.manage_addIndex(id, 'UUIDIndex', extra=extra,
                                REQUEST=REQUEST, RESPONSE=RESPONSE, URL1=URL3)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import random
import unittest
from itertools import chain

import ExtensionClass
from BTrees.IIBTree import IISet

from Products.PluginIndexes.FieldIndex.FieldIndex import FieldIndex
from Products.PluginIndexes.KeywordIndex.KeywordIndex import KeywordIndex
from Products.ZCTextIndex.OkapiIndex import OkapiIndex
from Products.ZCTextIndex.ZCTextIndex import PLexicon
from Products.ZCTextIndex.ZCTextIndex import ZCTextIndex


class ZDummy(ExtensionClass.Base):
    def __init__(self, num):
        self.num = num

    def title(self):
        return '{0:d}'.format(self.num)


class Dummy(ExtensionClass.Base):

    att1 = 'att1'
    att2 = 'att2'
    att3 = ['att3']
    foo = 'foo'

    def __init__(self, num):
        self.num = num
        if isinstance(num, int) and (self.num % 10) == 0:
            self.ends_in_zero = True

    def col1(self):
        return 'col1'

    def col2(self):
        return 'col2'

    def col3(self):
        return ['col3']


class MultiFieldIndex(FieldIndex):

    def getIndexQueryNames(self):
        return [self.id, 'bar']


class ObjRS(ExtensionClass.Base):

    def __init__(self, num):
        self.number = num


class TestAddDelColumn(unittest.TestCase):

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        return Catalog()

    def test_add(self):
        catalog = self._make_one()
        catalog.addColumn('id')
        self.assertIn('id', catalog.schema, 'add column failed')

    def test_add_bad(self):
        from Products.ZCatalog.Catalog import CatalogError
        catalog = self._make_one()
        self.assertRaises(CatalogError, catalog.addColumn, '_id')

    def test_add_with_space(self):
        catalog = self._make_one()
        catalog.addColumn(' space ')
        self.assertNotIn(' space ', catalog.schema,
                         'space not stripped in add column')
        self.assertIn('space', catalog.schema,
                      'stripping space in add column failed')

    def test_add_brains(self):
        catalog = self._make_one()
        catalog.addColumn('col1')
        catalog.addColumn('col3')
        for i in range(3):
            catalog.catalogObject(Dummy(3), repr(i))
        self.assertNotIn('col2', catalog.data.values()[0])
        catalog.addColumn('col2', default_value='new')
        self.assertIn('col2', catalog.schema, 'add column failed')
        self.assertIn('new', catalog.data.values()[0])

    def test_add_threshold(self):
        catalog = self._make_one()
        catalog.addColumn('col1')
        for i in range(3):
            catalog.catalogObject(Dummy(3), repr(i))
        catalog.addColumn('col2', threshold=None)

    def test_del(self):
        catalog = self._make_one()
        catalog.addColumn('id')
        catalog.delColumn('id')
        self.assertNotIn('id', catalog.schema, 'del column failed')

    def test_del_brains(self):
        catalog = self._make_one()
        catalog.addColumn('col1')
        catalog.addColumn('col2')
        catalog.addColumn('col3')
        for i in range(3):
            catalog.catalogObject(Dummy(3), repr(i))
        self.assertIn('col2', catalog.data.values()[0])
        catalog.delColumn('col2')
        self.assertNotIn('col2', catalog.schema, 'del column failed')
        self.assertNotIn('col2', catalog.data.values()[0])

    def test_del_threshold(self):
        catalog = self._make_one()
        catalog.addColumn('col1')
        for i in range(3):
            catalog.catalogObject(Dummy(3), repr(i))
        catalog.delColumn('col1', threshold=None)


class TestAddDelIndexes(unittest.TestCase):

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        return Catalog()

    def test_add_field_index(self):
        catalog = self._make_one()
        idx = FieldIndex('id')
        catalog.addIndex('id', idx)
        self.assertIsInstance(catalog.indexes['id'], FieldIndex)

    def test_add_text_index(self):
        catalog = self._make_one()
        catalog.lexicon = PLexicon('lexicon')
        idx = ZCTextIndex('id', caller=catalog,
                          index_factory=OkapiIndex, lexicon_id='lexicon')
        catalog.addIndex('id', idx)
        i = catalog.indexes['id']
        self.assertIsInstance(i, ZCTextIndex)

    def test_add_keyword_index(self):
        catalog = self._make_one()
        idx = KeywordIndex('id')
        catalog.addIndex('id', idx)
        i = catalog.indexes['id']
        self.assertIsInstance(i, KeywordIndex)

    def test_add_with_space(self):
        catalog = self._make_one()
        idx = KeywordIndex(' space ')
        catalog.addIndex(' space ', idx)
        self.assertNotIn(' space ', catalog.indexes,
                         'space not stripped in add index')
        self.assertIn('space', catalog.indexes,
                      'stripping space in add index failed')
        i = catalog.indexes['space']
        # Note: i.id still has spaces in it.
        self.assertIsInstance(i, KeywordIndex)

    def test_del_field_index(self):
        catalog = self._make_one()
        idx = FieldIndex('id')
        catalog.addIndex('id', idx)
        catalog.delIndex('id')
        self.assertNotIn('id', catalog.indexes)

    def test_del_text_index(self):
        catalog = self._make_one()
        catalog.lexicon = PLexicon('lexicon')
        idx = ZCTextIndex('id', caller=catalog,
                          index_factory=OkapiIndex, lexicon_id='lexicon')
        catalog.addIndex('id', idx)
        catalog.delIndex('id')
        self.assertNotIn('id', catalog.indexes)

    def test_del_keyword_index(self):
        catalog = self._make_one()
        idx = KeywordIndex('id')
        catalog.addIndex('id', idx)
        catalog.delIndex('id')
        self.assertNotIn('id', catalog.indexes)


class TestCatalog(unittest.TestCase):

    upper = 10

    nums = list(range(upper))
    for i in range(upper):
        j = random.randrange(0, upper)
        tmp = nums[i]
        nums[i] = nums[j]
        nums[j] = tmp

    def _make_one(self, extra=None):
        from Products.ZCatalog.Catalog import Catalog
        catalog = Catalog()
        catalog.lexicon = PLexicon('lexicon')

        att1 = FieldIndex('att1')
        catalog.addIndex('att1', att1)

        att2 = ZCTextIndex('att2', caller=catalog,
                           index_factory=OkapiIndex, lexicon_id='lexicon')
        catalog.addIndex('att2', att2)

        att3 = KeywordIndex('att3')
        catalog.addIndex('att3', att3)

        if extra is not None:
            extra(catalog)

        for x in range(0, self.upper):
            catalog.catalogObject(Dummy(self.nums[x]), repr(x))
        return catalog.__of__(Dummy('foo'))

    def test_clear(self):
        catalog = self._make_one()
        self.assertTrue(len(catalog) > 0)
        catalog.clear()
        self.assertEqual(catalog._length(), 0)
        self.assertEqual(len(catalog), 0)
        self.assertEqual(len(catalog.data), 0)
        self.assertEqual(len(catalog.paths), 0)
        self.assertEqual(len(catalog.uids), 0)
        for index_id in catalog.indexes:
            index = catalog.getIndex(index_id)
            self.assertEqual(index.numObjects(), 0)

    def test_getitem(self):
        def extra(catalog):
            catalog.addColumn('att1')

        catalog = self._make_one(extra=extra)
        catalog_rids = set(catalog.data)
        brain_class = catalog._v_result_class
        brains = []
        brain_rids = set()
        for rid in catalog_rids:
            brain = catalog[rid]
            brains.append(brain)
            brain_rids.add(brain.getRID())
            self.assertIsInstance(brain, brain_class)
            self.assertEqual(brain.att1, 'att1')

        self.assertEqual(len(brains), len(catalog))
        self.assertEqual(catalog_rids, brain_rids)

    # updateBrains
    # __setstate__
    # useBrains
    # getIndex
    # updateMetadata

    def testCatalogObjectUpdateMetadataFalse(self):
        def extra(catalog):
            catalog.addColumn('att1')
            num = FieldIndex('num')
            catalog.addIndex('num', num)

        catalog = self._make_one(extra=extra)
        ob = Dummy(9999)
        catalog.catalogObject(ob, '9999')
        brain = catalog(num=9999)[0]
        self.assertEqual(brain.att1, 'att1')
        ob.att1 = 'foobar'
        catalog.catalogObject(ob, '9999', update_metadata=0)
        brain = catalog(num=9999)[0]
        self.assertEqual(brain.att1, 'att1')
        catalog.catalogObject(ob, '9999')
        brain = catalog(num=9999)[0]
        self.assertEqual(brain.att1, 'foobar')

    def testUniqueValuesForLength(self):
        catalog = self._make_one()
        a = catalog.uniqueValuesFor('att1')
        self.assertEqual(
            len(a), 1,
            'bad number of unique values {0}'.format(a)
        )

    def testUniqueValuesForContent(self):
        catalog = self._make_one()
        a = catalog.uniqueValuesFor('att1')
        self.assertEqual(a[0], 'att1', 'bad content {0}'.format(a[0]))

    # hasuid
    # recordify
    # instantiate
    # getMetadataForRID
    # getIndexDataForRID
    # make_query

    def testKeywordIndexWithMinRange(self):
        catalog = self._make_one()
        a = catalog(att3={'query': 'att', 'range': 'min'})
        self.assertEqual(len(a), self.upper)

    def testKeywordIndexWithMaxRange(self):
        catalog = self._make_one()
        a = catalog(att3={'query': 'att35', 'range': ':max'})
        self.assertEqual(len(a), self.upper)

    def testKeywordIndexWithMinMaxRangeCorrectSyntax(self):
        catalog = self._make_one()
        a = catalog(att3={'query': ['att', 'att35'], 'range': 'min:max'})
        self.assertEqual(len(a), self.upper)

    def testKeywordIndexWithMinMaxRangeWrongSyntax(self):
        # checkKeywordIndex with min/max range wrong syntax.
        catalog = self._make_one()
        a = catalog(att3={'query': ['att'], 'range': 'min:max'})
        self.assertNotEqual(len(a), self.upper)

    def testCombinedTextandKeywordQuery(self):
        catalog = self._make_one()
        a = catalog(att3='att3', att2='att2')
        self.assertEqual(len(a), self.upper)
        a = catalog(att3='att3', att2='none')
        self.assertEqual(len(a), 0)

    def testResultLength(self):
        catalog = self._make_one()
        a = catalog(att1='att1')
        self.assertEqual(
            len(a), self.upper,
            'length should be {0}, its {1}'.format(self.upper, len(a))
        )

    def test_query_empty(self):
        # Queries with empty mappings used to return all.
        def extra(catalog):
            col1 = FieldIndex('col1')
            catalog.addIndex('col1', col1)
        catalog = self._make_one(extra=extra)
        self.assertTrue(len(catalog) > 0)
        all_data = catalog({})
        self.assertEqual(len(all_data), 0)

    def test_query_empty_keys(self):
        # Queries with empty keys used to return all.
        def extra(catalog):
            col1 = FieldIndex('col1')
            catalog.addIndex('col1', col1)
        catalog = self._make_one(extra=extra)
        a = catalog({'col1': ''})
        self.assertEqual(
            len(a), 0,
            'length should be 0, its {0}'.format(len(a))
        )

    def test_field_index_length(self):
        catalog = self._make_one()
        a = catalog(att1='att1')
        self.assertEqual(len(a), self.upper)
        a = catalog(att1='none')
        self.assertEqual(len(a), 0)

    def test_text_index_length(self):
        catalog = self._make_one()
        a = catalog(att2='att2')
        self.assertEqual(len(a), self.upper)
        a = catalog(att2='none')
        self.assertEqual(len(a), 0)

    def test_keyword_index_length(self):
        catalog = self._make_one()
        a = catalog(att3='att3')
        self.assertEqual(len(a), self.upper)
        a = catalog(att3='none')
        self.assertEqual(len(a), 0)


class TestCatalogSortBatch(unittest.TestCase):

    upper = 100

    nums = list(range(upper))
    for i in range(upper):
        j = random.randrange(0, upper)
        tmp = nums[i]
        nums[i] = nums[j]
        nums[j] = tmp

    def _make_one(self, extra=None):
        from Products.ZCatalog.Catalog import Catalog
        catalog = Catalog()
        catalog.lexicon = PLexicon('lexicon')
        att1 = FieldIndex('att1')
        att2 = ZCTextIndex('att2', caller=catalog,
                           index_factory=OkapiIndex, lexicon_id='lexicon')
        catalog.addIndex('att2', att2)
        num = FieldIndex('num')

        catalog.addIndex('att1', att1)
        catalog.addIndex('num', num)
        catalog.addColumn('num')

        foo = MultiFieldIndex('foo')
        catalog.addIndex('foo', foo)

        if extra is not None:
            extra(catalog)

        for x in range(0, self.upper):
            catalog.catalogObject(Dummy(self.nums[x]), repr(x))
        return catalog.__of__(Dummy('foo'))

    def test_sorted_search_indexes_empty(self):
        catalog = self._make_one()
        result = catalog._sorted_search_indexes({})
        self.assertEqual(len(result), 0)

    def test_sorted_search_indexes_one(self):
        catalog = self._make_one()
        result = catalog._sorted_search_indexes({'att1': 'a'})
        self.assertEqual(result, ['att1'])

    def test_sorted_search_indexes_many(self):
        catalog = self._make_one()
        query = {'att1': 'a', 'att2': 'b', 'num': 1}
        result = catalog._sorted_search_indexes(query)
        self.assertEqual(set(result), set(['att1', 'att2', 'num']))

    def test_sorted_search_indexes_priority(self):
        # att2 and col2 don't support ILimitedResultIndex, att1 does
        catalog = self._make_one()
        query = {'att1': 'a', 'att2': 'b', 'col2': 'c'}
        result = catalog._sorted_search_indexes(query)
        self.assertEqual(result.index('att2'), 0)
        self.assertEqual(result.index('att1'), 1)

    def test_sorted_search_indexes_match_alternate_attr(self):
        catalog = self._make_one()
        query = {'bar': 'b'}
        result = catalog._sorted_search_indexes(query)
        self.assertEqual(result, ['foo'])

    def test_sorted_search_indexes_no_match(self):
        catalog = self._make_one()
        result = catalog._sorted_search_indexes({'baz': 'a'})
        self.assertEqual(result, [])

    def test_sortResults(self):
        catalog = self._make_one()
        brains = catalog({'att1': 'att1'})
        rs = IISet([b.getRID() for b in brains])
        si = catalog.getIndex('num')
        result = catalog.sortResults(rs, si)
        self.assertEqual([r.num for r in result], list(range(100)))

    def test_sortResults_reversed(self):
        catalog = self._make_one()
        brains = catalog({'att1': 'att1'})
        rs = IISet([b.getRID() for b in brains])
        si = catalog.getIndex('num')
        result = catalog.sortResults(rs, si, reverse=True)
        self.assertEqual([r.num for r in result], list(reversed(range(100))))

    def test_sortResults_limit(self):
        catalog = self._make_one()
        brains = catalog({'att1': 'att1'})
        rs = IISet([b.getRID() for b in brains])
        si = catalog.getIndex('num')
        result = catalog.sortResults(rs, si, limit=10)
        self.assertEqual(len(result), 10)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(10)))

    def test_sortResults_limit_reversed(self):
        catalog = self._make_one()
        brains = catalog({'att1': 'att1'})
        rs = IISet([b.getRID() for b in brains])
        si = catalog.getIndex('num')
        result = catalog.sortResults(rs, si, reverse=True, limit=10)
        self.assertEqual(len(result), 10)
        self.assertEqual(result.actual_result_count, 100)
        expected = list(reversed(range(90, 100)))
        self.assertEqual([r.num for r in result], expected)

    def testLargeSortedResultSetWithSmallIndex(self):
        # This exercises the optimization in the catalog that iterates
        # over the sort index rather than the result set when the result
        # set is much larger than the sort index.
        catalog = self._make_one()
        a = catalog(att1='att1', sort_on='att1')
        self.assertEqual(len(a), self.upper)
        self.assertEqual(a.actual_result_count, self.upper)

    def testSortLimit(self):
        catalog = self._make_one()
        full = list(catalog(att1='att1', sort_on='num'))
        a = catalog(att1='att1', sort_on='num', sort_limit=10)
        self.assertEqual([r.num for r in a], [r.num for r in full[:10]])
        self.assertEqual(a.actual_result_count, self.upper)
        a = catalog(att1='att1', sort_on='num',
                    sort_limit=10, sort_order='reverse')
        rev = [r.num for r in full[-10:]]
        rev.reverse()
        self.assertEqual([r.num for r in a], rev)
        self.assertEqual(a.actual_result_count, self.upper)

    def testBigSortLimit(self):
        catalog = self._make_one()
        a = catalog(
            att1='att1', sort_on='num', sort_limit=self.upper * 3)
        self.assertEqual(a.actual_result_count, self.upper)
        self.assertEqual(a[0].num, 0)
        a = catalog(att1='att1', sort_on='num',
                    sort_limit=self.upper * 3, sort_order='reverse')
        self.assertEqual(a.actual_result_count, self.upper)
        self.assertEqual(a[0].num, self.upper - 1)

    def testSortLimitViaBatchingArgsBeforeStart(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=-5, b_size=8)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(0, 3)))

    def testSortLimitViaBatchingArgsStart(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=0, b_size=5)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(0, 5)))

    def testSortLimitViaBatchingEarlyFirstHalf(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=11, b_size=17)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(11, 28)))

    def testSortLimitViaBatchingArgsLateFirstHalf(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=30, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(30, 45)))

    def testSortLimitViaBatchingArgsLeftMiddle(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=45, b_size=8)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(45, 53)))

    def testSortLimitViaBatchingArgsRightMiddle(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=48, b_size=8)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(48, 56)))

    def testSortLimitViaBatchingArgsRightMiddleSortOnTwoSecond(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on=('att1', 'num'),
                     sort_order=('', 'reverse'), b_start=48, b_size=8)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(51, 43, -1)))

    def testSortLimitViaBatchingArgsEarlySecondHalf(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=55, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(55, 70)))

    def testSortLimitViaBatchingArgsEarlySecondHalfSortOnTwoFirst(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on=('att1', 'num'),
                     sort_order=('reverse', ''), b_start=55, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(55, 70)))

    def testSortLimitViaBatchingArgsEarlySecondHalfSortOnTwoSecond(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on=('att1', 'num'),
                     sort_order=('', 'reverse'), b_start=55, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(44, 29, -1)))

    def testSortLimitViaBatchingArgsEarlySecondHalfSortOnTwoBoth(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on=('att1', 'num'),
                     sort_order=('reverse', 'reverse'), b_start=55, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(44, 29, -1)))

    def testSortLimitViaBatchingArgsSecondHalf(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=70, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(70, 85)))

    def testSortLimitViaBatchingArgsEnd(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=90, b_size=10)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(90, 100)))

    def testSortLimitViaBatchingArgsOverEnd(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=90, b_size=15)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], list(range(90, 100)))

    def testSortLimitViaBatchingArgsOutside(self):
        catalog = self._make_one()
        query = dict(att1='att1', sort_on='num', b_start=110, b_size=10)
        result = catalog(query)
        self.assertEqual(result.actual_result_count, 100)
        self.assertEqual([r.num for r in result], [])

    def testSortedResultLengthWithMissingDocs(self):
        catalog = self._make_one()
        # remove the `0` document from the num index only
        num_index = catalog.getIndex('num')
        pos_of_zero = self.nums.index(0)
        uid = catalog.uids.get(repr(pos_of_zero))
        self.assertEqual(catalog[uid].num, 0)
        num_index.unindex_object(uid)
        # make sure it was removed
        self.assertEqual(len(num_index), 99)
        # sort over the smaller num index
        query = dict(att1='att1', sort_on='num', sort_limit=10)
        result = catalog(query)
        # the `0` document was removed
        self.assertEqual(result[0].num, 1)
        self.assertEqual(len(result), 10)
        # there are only 99 documents left
        self.assertEqual(result.actual_result_count, 99)

    # _get_sort_attr
    # _getSortIndex

    def test_search_not(self):
        catalog = self._make_one()
        query = dict(att1='att1', num={'not': [0, 1]})
        result = catalog(query)
        self.assertEqual(len(result), self.upper - 2)

    def test_search_not_nothing(self):
        def extra(catalog):
            col1 = FieldIndex('col1')
            catalog.addIndex('col1', col1)
        catalog = self._make_one(extra)
        query = dict(att1='att1', col1={'not': 'col1'})
        result = catalog(query)
        self.assertEqual(len(result), 0)

    def test_search_not_no_value_in_index(self):
        def extra(catalog):
            ends_in_zero = FieldIndex('ends_in_zero')
            catalog.addIndex('ends_in_zero', ends_in_zero)
        catalog = self._make_one(extra=extra)
        query = dict(att1='att1', ends_in_zero={'not': False})
        result = catalog(query)
        self.assertEqual(len(result), 10)

    def test_sort_on_good_index(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(att1='att1', sort_on='num')
        self.assertEqual(len(a), upper)
        for x in range(self.upper):
            self.assertEqual(a[x].num, x)

    def test_sort_on_bad_index(self):
        from Products.ZCatalog.Catalog import CatalogError
        catalog = self._make_one()

        def badsortindex():
            catalog(sort_on='foofaraw')
        self.assertRaises(CatalogError, badsortindex)

    def test_sort_on_wrong_index(self):
        from Products.ZCatalog.Catalog import CatalogError
        catalog = self._make_one()

        def wrongsortindex():
            catalog(sort_on='att2')
        self.assertRaises(CatalogError, wrongsortindex)

    def test_sort_on(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on='num', att2='att2')
        self.assertEqual(len(a), upper)
        for x in range(self.upper):
            self.assertEqual(a[x].num, x)

    def test_sort_on_missing(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(att2='att2')
        self.assertEqual(len(a), upper)

    def test_sort_on_two(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1')
        self.assertEqual(len(a), upper)
        for x in range(self.upper):
            self.assertEqual(a[x].num, x)

    def test_sort_on_two_reverse(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order='reverse')
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_two_reverse_with_limit(self):
        catalog = self._make_one()
        for num in range(-10, 0):
            obj = Dummy(num)
            obj.att1 = "att1foo"
            obj.att2 = "att2"
            catalog.catalogObject(obj, repr(num))
        a = catalog(
            att2='att2',
            sort_on=('att1', 'num'),
            sort_order='reverse',
            sort_limit=20,
        )
        self.assertEqual(
            [x.num for x in a[:10]],
            [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10],
        )
        self.assertEqual(
            [x.num for x in a[10:]],
            [99, 98, 97, 96, 95, 94, 93, 92, 91, 90],
        )

    def test_sort_on_two_reverse_neither(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order=('', ''))
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_two_reverse_first(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order=('reverse', ''))
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_two_reverse_second(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order=('', 'reverse'))
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_two_reverse_both(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order=('reverse', 'reverse'))
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_two_reverse_too_many(self):
        catalog = self._make_one()
        upper = self.upper
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_order=('', '', 'reverse', ''))
        self.assertEqual(len(a), upper)
        for x in range(upper - 1):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_two_small_limit(self):
        catalog = self._make_one()
        a = catalog(sort_on=('att1', 'num'), att1='att1', sort_limit=10)
        self.assertEqual(len(a), 10)
        for x in range(9):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_two_small_limit_reverse(self):
        catalog = self._make_one()
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_limit=10, sort_order='reverse')
        self.assertEqual(len(a), 10)
        for x in range(9):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_two_big_limit(self):
        catalog = self._make_one()
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_limit=self.upper * 3)
        self.assertEqual(len(a), 100)
        for x in range(99):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_two_big_limit_reverse(self):
        catalog = self._make_one()
        a = catalog(sort_on=('att1', 'num'), att1='att1',
                    sort_limit=self.upper * 3, sort_order='reverse')
        self.assertEqual(len(a), 100)
        for x in range(99):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_three(self):
        def extra(catalog):
            col2 = FieldIndex('col2')
            catalog.addIndex('col2', col2)
        catalog = self._make_one(extra)
        a = catalog(sort_on=('att1', 'col2', 'num'), att1='att1')
        self.assertEqual(len(a), self.upper)
        for x in range(self.upper):
            self.assertEqual(a[x].num, x)

    def test_sort_on_three_reverse(self):
        def extra(catalog):
            col2 = FieldIndex('col2')
            catalog.addIndex('col2', col2)
        catalog = self._make_one(extra)
        a = catalog(sort_on=('att1', 'col2', 'num'), att1='att1',
                    sort_order='reverse')
        self.assertEqual(len(a), self.upper)
        for x in range(self.upper - 1):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_three_reverse_last(self):
        def extra(catalog):
            col2 = FieldIndex('col2')
            catalog.addIndex('col2', col2)
        catalog = self._make_one(extra)
        a = catalog(sort_on=('att1', 'col2', 'num'), att1='att1',
                    sort_order=('', '', 'reverse'))
        self.assertEqual(len(a), self.upper)
        for x in range(self.upper - 1):
            self.assertTrue(a[x].num > a[x + 1].num)

    def test_sort_on_three_small_limit(self):
        def extra(catalog):
            col2 = FieldIndex('col2')
            catalog.addIndex('col2', col2)
        catalog = self._make_one(extra)
        a = catalog(sort_on=('att1', 'col2', 'num'), att1='att1',
                    sort_limit=10)
        self.assertEqual(len(a), 10)
        for x in range(9):
            self.assertTrue(a[x].num < a[x + 1].num)

    def test_sort_on_three_big_limit(self):
        def extra(catalog):
            col2 = FieldIndex('col2')
            catalog.addIndex('col2', col2)
        catalog = self._make_one(extra)
        a = catalog(sort_on=('att1', 'col2', 'num'), att1='att1',
                    sort_limit=self.upper * 3)
        self.assertEqual(len(a), 100)
        for x in range(99):
            self.assertTrue(a[x].num < a[x + 1].num)


class TestUnCatalog(unittest.TestCase):

    upper = 5

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        catalog = Catalog()
        catalog.lexicon = PLexicon('lexicon')
        att1 = FieldIndex('att1')
        att2 = ZCTextIndex('att2', caller=catalog,
                           index_factory=OkapiIndex, lexicon_id='lexicon')
        att3 = KeywordIndex('att3')
        catalog.addIndex('att1', att1)
        catalog.addIndex('att2', att2)
        catalog.addIndex('att3', att3)

        for x in range(0, self.upper):
            catalog.catalogObject(Dummy(x), repr(x))
        return catalog.__of__(Dummy('foo'))

    def _uncatalog(self, catalog):
        for x in range(0, self.upper):
            catalog.uncatalogObject(repr(x))

    def test_uncatalog_field_index(self):
        catalog = self._make_one()
        self._uncatalog(catalog)
        a = catalog(att1='att1')
        self.assertEqual(len(a), 0, 'len: {0}'.format(len(a)))

    def test_uncatalog_text_index(self):
        catalog = self._make_one()
        self._uncatalog(catalog)
        a = catalog(att2='att2')
        self.assertEqual(len(a), 0, 'len: {0}'.format(len(a)))

    def test_uncatalog_keyword_index(self):
        catalog = self._make_one()
        self._uncatalog(catalog)
        a = catalog(att3='att3')
        self.assertEqual(len(a), 0, 'len: {0}'.format(len(a)))

    def test_bad_uncatalog(self):
        catalog = self._make_one()
        try:
            catalog.uncatalogObject('asdasdasd')
        except Exception:
            self.fail('uncatalogObject raised exception on bad uid')

    def test_uncatalog_twice(self):
        catalog = self._make_one()
        catalog.uncatalogObject('0')

        def _second(self):
            catalog.uncatalogObject('0')
        self.assertRaises(Exception, _second)

    def test_uncatalog_ength(self):
        catalog = self._make_one()
        self._uncatalog(catalog)
        self.assertEqual(len(catalog), 0)


class TestRangeSearch(unittest.TestCase):

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        return Catalog()

    def test_range_search(self):
        catalog = self._make_one()
        index = FieldIndex('number')
        catalog.addIndex('number', index)
        catalog.addColumn('number')
        for i in range(50):
            obj = ObjRS(random.randrange(0, 200))
            catalog.catalogObject(obj, i)
        catalog = catalog.__of__(ObjRS(20))

        for i in range(10):
            m = random.randrange(0, 200)
            n = m + 10
            for r in catalog(number={'query': (m, n), 'range': 'min:max'}):
                size = r.number
                self.assertTrue(
                    m <= size and size <= n,
                    '{0:d} vs [{1:d},{2:d}]'.format(r.number, m, n)
                )


class TestMergeResults(unittest.TestCase):

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        return Catalog()

    def _make_many(self):
        from Products.ZCatalog.Catalog import mergeResults
        catalogs = []
        for i in range(3):
            cat = self._make_one()
            cat.lexicon = PLexicon('lexicon')
            cat.addIndex('num', FieldIndex('num'))
            cat.addIndex('big', FieldIndex('big'))
            cat.addIndex('number', FieldIndex('number'))
            i = ZCTextIndex('title', caller=cat, index_factory=OkapiIndex,
                            lexicon_id='lexicon')
            cat.addIndex('title', i)
            cat = cat.__of__(ZDummy(16336))
            for i in range(10):
                obj = ZDummy(i)
                obj.big = i > 5
                obj.number = True
                cat.catalogObject(obj, str(i))
            catalogs.append(cat)
        return catalogs, mergeResults

    def _sort(self, iterable, reverse=False):
        L = list(iterable)
        if reverse:
            L.sort(reverse=True)
        else:
            L.sort()
        return L

    def test_no_filter_or_sort(self):
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(
                   dict(number=True), _merge=0) for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=False, reverse=False)]
        expected = [r.getRID() for r in chain(*results)]
        self.assertEqual(self._sort(merged_rids), self._sort(expected))

    def test_sorted_only(self):
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(
                   dict(number=True, sort_on='num'), _merge=0)
                   for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=True, reverse=False)]
        expected = self._sort(chain(*results))
        expected = [rid for sortkey, rid, getitem in expected]
        self.assertEqual(merged_rids, expected)

    def test_sort_reverse(self):
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(
                   dict(number=True, sort_on='num'), _merge=0)
                   for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=True, reverse=True)]
        expected = self._sort(chain(*results), reverse=True)
        expected = [rid for sortkey, rid, getitem in expected]
        self.assertEqual(merged_rids, expected)

    def test_limit_sort(self):
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(
                   dict(att1='att1', number=True, sort_on='num',
                        sort_limit=2), _merge=0)
                   for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=True, reverse=False)]
        expected = self._sort(chain(*results))
        expected = [rid for sortkey, rid, getitem in expected]
        self.assertEqual(merged_rids, expected)

    def test_scored(self):
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(title='4 or 5 or 6', _merge=0)
                   for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=True, reverse=False)]
        expected = self._sort(chain(*results))
        expected = [rid for sortkey, (nscore, score, rid), getitem in expected]
        self.assertEqual(merged_rids, expected)

    def test_small_index_sort(self):
        # Test that small index sort optimization is not used for merging
        catalogs, mergeResults = self._make_many()
        results = [cat.searchResults(
                   dict(number=True, sort_on='big'), _merge=0)
                   for cat in catalogs]
        merged_rids = [r.getRID() for r in mergeResults(
            results, has_sort_keys=True, reverse=False)]
        expected = self._sort(chain(*results))
        expected = [rid for sortkey, rid, getitem in expected]
        self.assertEqual(merged_rids, expected)


class TestScoring(unittest.TestCase):

    def _make_one(self):
        from Products.ZCatalog.Catalog import Catalog
        catalog = Catalog()
        catalog.lexicon = PLexicon('lexicon')
        idx = ZCTextIndex('title', caller=catalog,
                          index_factory=OkapiIndex, lexicon_id='lexicon')
        catalog.addIndex('title', idx)
        catalog.addIndex('true', FieldIndex('true'))
        catalog.addColumn('title')
        for i in (1, 2, 3, 10, 11, 110, 111):
            obj = ZDummy(i)
            obj.true = True
            if i == 110:
                obj.true = False
            catalog.catalogObject(obj, str(i))
        return catalog.__of__(ZDummy(1))

    def test_simple_search(self):
        cat = self._make_one()
        brains = cat(title='10')
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].title, '10')

    def test_or_search(self):
        cat = self._make_one()
        brains = cat(title='2 OR 3')
        self.assertEqual(len(brains), 2)

    def test_scored_search(self):
        cat = self._make_one()
        brains = cat(title='1*')
        self.assertEqual(len(brains), 5)
        self.assertEqual(brains[0].title, '111')

    def test_combined_scored_search(self):
        cat = self._make_one()
        brains = cat(title='1*', true=True)
        self.assertEqual(len(brains), 4)
        self.assertEqual(brains[0].title, '111')

    def test_combined_scored_search_planned(self):
        from ..plan import Benchmark
        from ..plan import PriorityMap
        cat = self._make_one()
        query = dict(title='1*', true=True)
        plan = cat.getCatalogPlan()
        plan_key = plan.make_key(query)
        catalog_id = plan.get_id()
        # plan with title first
        PriorityMap.set_entry(catalog_id, plan_key, dict(
            title=Benchmark(1, 1, False),
            true=Benchmark(2, 1, False),
        ))
        brains = cat(query)
        self.assertEqual(len(brains), 4)
        # plan with true first
        PriorityMap.set_entry(catalog_id, plan_key, dict(
            title=Benchmark(2, 1, False),
            true=Benchmark(1, 1, False),
        ))
        brains = cat(query)
        self.assertEqual(len(brains), 4)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest
from datetime import datetime

from BTrees.IIBTree import IISet
from DateTime.DateTime import DateTime
from OFS.SimpleItem import SimpleItem
from Testing.makerequest import makerequest

from Products.PluginIndexes.util import datetime_to_minutes


class Dummy(object):

    def __init__(self, name, start, stop):
        self._name = name
        self._start = start
        self._stop = stop

    def name(self):
        return self._name

    def start(self):
        return self._start

    def stop(self):
        return self._stop

    def datum(self):
        return (self._start, self._stop)


dummies = [(0, Dummy('a', None, None)),
           (1, Dummy('b', None, None)),
           (2, Dummy('c', 0, None)),
           (3, Dummy('d', 10, None)),
           (4, Dummy('e', None, 4)),
           (5, Dummy('f', None, 11)),
           (6, Dummy('g', 0, 11)),
           (7, Dummy('h', 2, 9)),
           ]


def matchingDummiesByTimeValue(value, precision=1):
    result = []
    value = datetime_to_minutes(value, precision)
    for i, dummy in dummies:
        start = datetime_to_minutes(dummy.start(), precision)
        stop = datetime_to_minutes(dummy.stop(), precision)
        if ((start is None or start <= value)
                and (stop is None or stop >= value)):
            result.append((i, dummy))
    return result


def matchingDummiesByUIDs(uids):
    return [(i, dummies[i]) for i in uids]


class DateRangeIndexTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.PluginIndexes.DateRangeIndex.DateRangeIndex import \
            DateRangeIndex
        return DateRangeIndex

    def _makeOne(self, id, since_field=None, until_field=None,
                 caller=None, extra=None, precision_value=None):
        klass = self._getTargetClass()
        index = klass(id, since_field, until_field, caller, extra,
                      precision_value=precision_value)

        class DummyZCatalog(SimpleItem):
            id = 'DummyZCatalog'

        # Build pseudo catalog and REQUEST environment
        catalog = makerequest(DummyZCatalog())
        indexes = SimpleItem()

        indexes = indexes.__of__(catalog)
        index = index.__of__(indexes)

        return index

    def _checkApply(self, index, req, expectedValues, resultset=None):

        def checkApply():
            result, used = index._apply_index(req, resultset=resultset)
            try:
                result = result.keys()
            except AttributeError:
                pass

            assert used == (index._since_field, index._until_field)
            assert len(result) == len(expectedValues), \
                '{0}: {1} | {2}'.format(req, list(result), expectedValues)
            for k, v in expectedValues:
                assert k in result
            return (result, used)

        cache = index.getRequestCache()
        cache.clear()

        # first call; regular check
        result, used = checkApply()
        self.assertEqual(cache._hits, 0)
        self.assertEqual(cache._sets, 1)
        self.assertEqual(cache._misses, 1)

        # second call; caching check
        result, used = checkApply()
        self.assertEqual(cache._hits, 1)
        return result, used

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from Products.PluginIndexes.interfaces import IDateRangeIndex
        from Products.PluginIndexes.interfaces import IPluggableIndex
        from Products.PluginIndexes.interfaces import ISortIndex
        from Products.PluginIndexes.interfaces import IUniqueValueIndex

        verifyClass(IDateRangeIndex, self._getTargetClass())
        verifyClass(IPluggableIndex, self._getTargetClass())
        verifyClass(ISortIndex, self._getTargetClass())
        verifyClass(IUniqueValueIndex, self._getTargetClass())

    def test_empty(self):
        empty = self._makeOne('empty')

        self.assertTrue(empty.getEntryForObject(1234) is None)
        empty.unindex_object(1234)  # shouldn't throw

        self.assertFalse(list(empty.uniqueValues('foo')))
        self.assertFalse(list(empty.uniqueValues('foo', withLengths=True)))
        self.assertTrue(empty._apply_index({'zed': 12345}) is None)

        self._checkApply(empty, {'empty': 12345}, [])

    def test_retrieval(self):
        index = self._makeOne('work', 'start', 'stop')

        for i, dummy in dummies:
            result = index.index_object(i, dummy)
            self.assertEqual(result, 1)

            # don't index datum twice
            result = index.index_object(i, dummy)
            self.assertEqual(result, 0)

        for i, dummy in dummies:
            self.assertEqual(index.getEntryForObject(i), dummy.datum())

        for value in range(-1, 15):
            matches = matchingDummiesByTimeValue(value)
            results, used = self._checkApply(index, {'work': value}, matches)
            matches = sorted(matches, key=lambda d: d[1].name())

            for result, match in zip(results, matches):
                self.assertEqual(index.getEntryForObject(result),
                                 match[1].datum())

        # check update
        i, dummy = dummies[0]
        start = dummy._start
        dummy._start = start and start + 1 or 1
        index.index_object(i, dummy)
        self.assertEqual(index.getEntryForObject(0), dummy.datum())

    def test_longdates(self):
        too_large = 2 ** 31
        too_small = -2 ** 31
        index = self._makeOne('work', 'start', 'stop')
        bad = Dummy('bad', too_large, too_large)
        self.assertRaises(OverflowError, index.index_object, 0, bad)
        bad = Dummy('bad', too_small, too_small)
        self.assertRaises(OverflowError, index.index_object, 0, bad)

    def test_floor_date(self):
        index = self._makeOne('work', 'start', 'stop')
        floor = index.floor_value - 1
        bad = Dummy('bad', floor, None)
        index.index_object(0, bad)
        self.assertTrue(0 in index._always.keys())

    def test_ceiling_date(self):
        index = self._makeOne('work', 'start', 'stop')
        ceiling = index.ceiling_value + 1
        bad = Dummy('bad', None, ceiling)
        index.index_object(1, bad)
        self.assertTrue(1 in index._always.keys())

    def test_datetime(self):
        from Products.PluginIndexes.DateIndex.tests import _getEastern
        before = datetime(2009, 7, 11, 0, 0, tzinfo=_getEastern())
        start = datetime(2009, 7, 13, 5, 15, tzinfo=_getEastern())
        between = datetime(2009, 7, 13, 5, 45, tzinfo=_getEastern())
        stop = datetime(2009, 7, 13, 6, 30, tzinfo=_getEastern())
        after = datetime(2009, 7, 14, 0, 0, tzinfo=_getEastern())

        dummy = Dummy('test', start, stop)
        index = self._makeOne('work', 'start', 'stop')
        index.index_object(0, dummy)
        matches = matchingDummiesByUIDs([0])

        self.assertEqual(index.getEntryForObject(0),
                         (DateTime(start).millis() / 60000,
                          DateTime(stop).millis() / 60000))

        self._checkApply(index, {'work': before}, [])

        self._checkApply(index, {'work': start}, matches)

        self._checkApply(index, {'work': between}, matches)

        self._checkApply(index, {'work': stop}, matches)

        self._checkApply(index, {'work': after}, [])

    def test_datetime_naive_timezone(self):
        from Products.PluginIndexes.DateIndex.DateIndex import Local
        before = datetime(2009, 7, 11, 0, 0)
        start = datetime(2009, 7, 13, 5, 15)
        start_local = datetime(2009, 7, 13, 5, 15, tzinfo=Local)
        between = datetime(2009, 7, 13, 5, 45)
        stop = datetime(2009, 7, 13, 6, 30)
        stop_local = datetime(2009, 7, 13, 6, 30, tzinfo=Local)
        after = datetime(2009, 7, 14, 0, 0)

        dummy = Dummy('test', start, stop)
        index = self._makeOne('work', 'start', 'stop')
        index.index_object(0, dummy)
        matches = matchingDummiesByUIDs([0])

        self.assertEqual(index.getEntryForObject(0),
                         (DateTime(start_local).millis() / 60000,
                          DateTime(stop_local).millis() / 60000))

        self._checkApply(index, {'work': before}, [])

        self._checkApply(index, {'work': start}, matches)

        self._checkApply(index, {'work': between}, matches)

        self._checkApply(index, {'work': stop}, matches)

        self._checkApply(index, {'work': after}, [])

    def test_resultset(self):
        index = self._makeOne('work', 'start', 'stop')
        for i, dummy in dummies:
            index.index_object(i, dummy)

        self._checkApply(index, {'work': 20},
                         matchingDummiesByUIDs([0, 1, 2, 3]))

        # a resultset with everything doesn't actually limit
        self._checkApply(index, {'work': 20},
                         matchingDummiesByUIDs([0, 1, 2, 3]),
                         resultset=IISet(range(len(dummies))))

        # a small resultset limits
        self._checkApply(index, {'work': 20},
                         matchingDummiesByUIDs([1, 2]),
                         resultset=IISet([1, 2]))

        # the specified value is included
        self._checkApply(index, {'work': 11},
                         matchingDummiesByUIDs([0, 1, 2, 3, 5, 6]))

        # also for _since_only
        self._checkApply(index, {'work': 10},
                         matchingDummiesByUIDs([0, 1, 2, 3, 5, 6]))

        # the specified value is included with a large resultset
        self._checkApply(index, {'work': 11},
                         matchingDummiesByUIDs([0, 1, 2, 3, 5, 6]),
                         resultset=IISet(range(len(dummies))))

        # this also works for _since_only
        self._checkApply(index, {'work': 10},
                         matchingDummiesByUIDs([0, 1, 2, 3, 5, 6]),
                         resultset=IISet(range(len(dummies))))

        # the specified value is included with a small resultset
        self._checkApply(index, {'work': 11},
                         matchingDummiesByUIDs([0, 5]),
                         resultset=IISet([0, 5, 7]))

    def test_getCounter(self):
        index = self._makeOne('work', 'start', 'stop')
        self.assertEqual(index.getCounter(), 0)

        k, obj = dummies[0]
        index.index_object(k, obj)
        self.assertEqual(index.getCounter(), 1)

        index.unindex_object(k)
        self.assertEqual(index.getCounter(), 2)

        # unknown id
        index.unindex_object(1234)
        self.assertEqual(index.getCounter(), 2)

        # clear is a change
        index.clear()
        self.assertEqual(index.getCounter(), 3)

    def test_precision(self):
        precision = 5
        index = self._makeOne('work', 'start', 'stop',
                              precision_value=precision)

        for i, dummy in dummies:
            index.index_object(i, dummy)

        for i, dummy in dummies:
            datum = map(lambda d: datetime_to_minutes(d, precision),
                        dummy.datum())
            self.assertEqual(index.getEntryForObject(i), tuple(datum))

        for value in range(-1, 15):
            matches = matchingDummiesByTimeValue(value, precision)
            results, used = self._checkApply(index, {'work': value}, matches)
            matches = sorted(matches, key=lambda d: d[1].name())
            for result, match in zip(results, matches):
                datum = map(lambda d: datetime_to_minutes(d, precision),
                            match[1].datum())
                self.assertEqual(index.getEntryForObject(result),
                                 tuple(datum))



import Testing.testbrowser
import Testing.ZopeTestCase
import Zope2.App


class IndexSortingTests(Testing.ZopeTestCase.FunctionalTestCase):
    """Browser testing /manage_catalogIndexes"""

    def setUp(self):
        super(IndexSortingTests, self).setUp()

        Zope2.App.zcml.load_site(force=True)

        uf = self.app.acl_users
        uf.userFolderAddUser('manager', 'manager_pass', ['Manager'], [])
        zcatalog = self.app.manage_addProduct['ZCatalog']
        zcatalog.manage_addZCatalog('catalog', 'The Catalog')
        pli = self.app.catalog.Indexes.manage_addProduct['PluginIndexes']
        pli.manage_addFieldIndex('Index1')
        pli.manage_addKeywordIndex('Index2')
        self.browser = Testing.testbrowser.Browser()
        self.browser.login('manager', 'manager_pass')

    def check_order(self, expect_1_before_2):
        index1_pos = self.browser.contents.find('Index1')
        index2_pos = self.browser.contents.find('Index2')
        found_1_before_2 = index2_pos > index1_pos
        self.assertEqual(found_1_before_2, expect_1_before_2)

    def test_sortby(self):
        base_url = (
            'http://localhost/catalog/manage_catalogIndexes?skey=%s&rkey=%s')

        self.browser.open(base_url % ('id', 'asc'))
        self.check_order(expect_1_before_2=True)

        self.browser.open(base_url % ('id', 'desc'))
        self.check_order(expect_1_before_2=False)

        self.browser.open(base_url % ('meta_type', 'asc'))
        self.check_order(expect_1_before_2=True)

        self.browser.open(base_url % ('meta_type', 'desc'))
        self.check_order(expect_1_before_2=False)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from logging import getLogger

import BTrees.Length
from App.special_dtml import DTMLFile
from BTrees.IIBTree import IIBTree
from BTrees.IIBTree import IISet
from BTrees.IIBTree import IITreeSet
from BTrees.IIBTree import difference
from BTrees.IIBTree import intersection
from BTrees.IIBTree import union
from ZODB.POSException import ConflictError

from Products.PluginIndexes.unindex import UnIndex
from Products.PluginIndexes.unindex import _marker


LOG = getLogger('BooleanIndex.UnIndex')


class BooleanIndex(UnIndex):
    """Index for booleans

       self._index = set([documentId1, documentId2])
       self._unindex = {documentId:[True/False]}

       self._length is the length of the unindex
       self._index_length is the length of the index
       self._index_value is the indexed value

       The document ids in self._index have self._index_value as their value.

       Since there are only two possible values (True/False), the index
       only stores a forward index for the less common value.
       It starts off with the opposite of value of the first document
       and later checks and inverts itself, if more than 60% of all
       documents now have the indexed value. It does the inversion
       at 60% to avoid inverting itself constantly for an index that
       has a roughly equal 50/50 split.
    """

    meta_type = 'BooleanIndex'

    manage_options = (
        {'label': 'Settings',
         'action': 'manage_main'},
        {'label': 'Browse',
         'action': 'manage_browse'},
    )

    query_options = ['query']

    manage = manage_main = DTMLFile('dtml/manageBooleanIndex', globals())
    manage_main._setName('manage_main')
    manage_browse = DTMLFile('../dtml/browseIndex', globals())

    _index_value = 1
    _index_length = None

    def clear(self):
        self._index = IITreeSet()
        self._index_length = BTrees.Length.Length()
        self._index_value = 1
        self._unindex = IIBTree()
        self._length = BTrees.Length.Length()
        if self._counter is None:
            self._counter = BTrees.Length.Length()
        else:
            self._increment_counter()

    def histogram(self):
        """Return a mapping which provides a histogram of the number of
        elements found at each point in the index.
        """
        histogram = {}
        indexed = bool(self._index_value)
        histogram[indexed] = self._index_length.value
        histogram[not indexed] = self._length.value - self._index_length.value
        return histogram

    def _invert_index(self, documentId=None):
        self._index_value = indexed = int(not self._index_value)
        self._index.clear()
        length = 0
        for rid, value in self._unindex.iteritems():
            if value == indexed:
                self._index.add(rid)
                length += 1
        # documentId is the rid of the currently processed object that
        # triggered the invert. in the case of unindexing, the rid hasn't
        # been removed from the unindex yet. While indexing, the rid will
        # be added to the index and unindex after this method is done
        if documentId is not None:
            self._index.remove(documentId)
            length -= 1
        self._index_length = BTrees.Length.Length(length)

    def insertForwardIndexEntry(self, entry, documentId):
        """If the value matches the indexed one, insert into treeset
        """
        # When we get the first entry, decide to index the opposite of what
        # we got, as indexing zero items is fewer than one.
        if self._length.value == 0:
            self._index_value = int(not bool(entry))

        # if the added entry value is index value, insert it into index
        if bool(entry) is bool(self._index_value):
            self._index_length.change(1)
            self._index.insert(documentId)

        # insert value into global unindex (before computing index invert)
        self._unindex[documentId] = entry
        self._length.change(1)

        # is the index (after adding the current entry) larger than 60%
        # of the total length? than switch the indexed value
        if bool(entry) is bool(self._index_value):
            if (self._index_length.value) >= ((self._length.value) * 0.6):
                self._invert_index()

    def removeForwardIndexEntry(self, entry, documentId, check=True):
        """Take the entry provided and remove any reference to documentId
        in its entry in the index.
        """
        if bool(entry) is bool(self._index_value):
            try:
                self._index.remove(documentId)
                self._index_length.change(-1)
            except ConflictError:
                raise
            except Exception:
                LOG.exception('%(context)s: unindex_object could not '
                              'remove documentId %(doc_id)s from '
                              'index %(index)r. This should not '
                              'happen.', dict(
                                  context=self.__class__.__name__,
                                  doc_id=documentId,
                                  index=self.id))
        elif check:
            # is the index (after removing the current entry) larger than
            # 60% of the total length? than switch the indexed value
            if (self._index_length.value) <= ((self._length.value - 1) * 0.6):
                self._invert_index(documentId)
                return

    def _index_object(self, documentId, obj, threshold=None, attr=''):
        """ index and object 'obj' with integer id 'documentId'"""
        returnStatus = 0

        # First we need to see if there's anything interesting to look at
        datum = self._get_object_datum(obj, attr)

        # Make it boolean, int as an optimization
        if datum is not _marker:
            datum = int(bool(datum))

        # We don't want to do anything that we don't have to here, so we'll
        # check to see if the new and existing information is the same.
        oldDatum = self._unindex.get(documentId, _marker)
        if datum != oldDatum:
            if oldDatum is not _marker:
                self.removeForwardIndexEntry(oldDatum, documentId, check=False)
                if datum is _marker:
                    try:
                        del self._unindex[documentId]
                        self._length.change(-1)
                    except ConflictError:
                        raise
                    except Exception:
                        LOG.error('Should not happen: oldDatum was there, now '
                                  'its not, for document with id %s',
                                  documentId)

            if datum is not _marker:
                self.insertForwardIndexEntry(datum, documentId)

            returnStatus = 1

        return returnStatus

    def unindex_object(self, documentId):
        """ Unindex the object with integer id 'documentId' and don't
        raise an exception if we fail
        """
        unindexRecord = self._unindex.get(documentId, _marker)
        if unindexRecord is _marker:
            return None

        self._increment_counter()

        self.removeForwardIndexEntry(unindexRecord, documentId)

        try:
            del self._unindex[documentId]
            self._length.change(-1)
        except ConflictError:
            raise
        except Exception:
            LOG.debug('Attempt to unindex nonexistent document'
                      ' with id %s', documentId, exc_info=True)

    def query_index(self, record, resultset=None):
        index = self._index
        indexed = self._index_value

        for key in record.keys:
            if bool(key) is bool(indexed):
                # If we match the indexed value, check index
                return intersection(index, resultset)
            else:
                # Otherwise, remove from resultset or _unindex
                if resultset is None:
                    return union(difference(self._unindex, index), IISet([]))
                else:
                    return intersection(difference(resultset, index),
                                        self._unindex)
        return IISet()

    def indexSize(self):
        """Return distinct values, as an optimization we always claim 2."""
        return 2

    def items(self):
        # return a list of value to int set of rid tuples
        indexed = self._index_value
        items = [(bool(indexed), self._index)]
        false = IISet()
        for rid, value in self._unindex.iteritems():
            if value != indexed:
                false.add(rid)
        items.append((not bool(indexed), false))
        return items

    def uniqueValues(self, name=None, withLengths=0):
        """returns the unique values for name

        if withLengths is true, returns a sequence of
        tuples of (value, length)
        """
        if name is None:
            name = self.id
        elif name != self.id:
            return

        indexed = bool(self._index_value)
        unique_values = (indexed, not indexed)
        if not withLengths:
            for key in unique_values:
                yield key
        else:
            for key in unique_values:
                ilen = len(self._index)
                if key is indexed:
                    yield (key, ilen)
                else:
                    ulen = len(self._unindex)
                    yield (key, ulen - ilen)


manage_addBooleanIndexForm = DTMLFile('dtml/addBooleanIndex', globals())


def manage_addBooleanIndex(self, id, extra=None,
                           REQUEST=None, RESPONSE=None, URL3=None):
    """Add a boolean index"""
    return self.manage_addIndex(
        id, 'BooleanIndex', extra=extra,
        REQUEST=REQUEST, RESPONSE=RESPONSE, URL1=URL3)



##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

B = 0.75
B_FROM1 = 0.25  # 1.0 - B

K1 = 1.2
K1_PLUS1 = 2.2  # K1 + 1.0


def score(result, d2fitems, d2len, idf, meandoclen):
    """
    Do the inner scoring loop for an Okapi index.

    result: IIBucket result, maps d to score
    d2fitems: _wordinfo[t].items(), maps d to f(d, t)
    d2len: _docweight, maps d to # words in d
    idf: inverse doc frequency of t
    meandoclen: average number of words in a doc

    BBB ZCTextIndex 4.0: Compatibility for former C implementation.
    """
    idf *= 1024.0  # float out part of the scaled_int computation

    for docid, f in d2fitems:
        f = int(f)
        lenweight = B_FROM1 + B * int(d2len[docid]) / meandoclen
        tf = f * K1_PLUS1 / (f + K1 * lenweight)
        result[docid] = int(tf * idf + 0.5)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Lexicon.
"""

import re
from random import randrange

import six

from BTrees.IOBTree import IOBTree
from BTrees.Length import Length
from BTrees.OIBTree import OIBTree
from Persistence import Persistent
from zope.interface import implementer

from Products.ZCTextIndex.interfaces import ILexicon
from Products.ZCTextIndex.ParseTree import QueryError
from Products.ZCTextIndex.PipelineFactory import element_factory
from Products.ZCTextIndex.StopDict import get_stopdict


@implementer(ILexicon)
class Lexicon(Persistent):

    _v_nextid = None
    _wid_length_based = True  # Flag to distinguish new and old lexica

    def __init__(self, *pipeline):
        self.clear()
        self._pipeline = pipeline

    def clear(self):
        """Empty the lexicon.
        """
        self.length = Length()
        self._wid_length_based = False
        self._wids = OIBTree()  # word -> wid
        self._words = IOBTree()  # wid -> word
        # wid 0 is reserved for words that aren't in the lexicon (OOV -- out
        # of vocabulary).  This can happen, e.g., if a query contains a word
        # we never saw before, and that isn't a known stopword (or otherwise
        # filtered out).  Returning a special wid value for OOV words is a
        # way to let clients know when an OOV word appears.

    def length(self):
        """Return the number of unique terms in the lexicon.
        """
        # Overridden in instances with a BTrees.Length.Length
        raise NotImplementedError

    def words(self):
        return self._wids.keys()

    def wids(self):
        return self._words.keys()

    def items(self):
        return self._wids.items()

    def sourceToWordIds(self, text):
        last = _text2list(text)
        for element in self._pipeline:
            last = element.process(last)
        return list(map(self._getWordIdCreate, last))

    def termToWordIds(self, text):
        last = _text2list(text)
        for element in self._pipeline:
            process = getattr(element, "process_post_glob", element.process)
            last = process(last)
        wids = []
        for word in last:
            wids.append(self._wids.get(word, 0))
        return wids

    def parseTerms(self, text):
        last = _text2list(text)
        for element in self._pipeline:
            process = getattr(element, "processGlob", element.process)
            last = process(last)
        return last

    def isGlob(self, word):
        return "*" in word or "?" in word

    def get_word(self, wid):
        return self._words[wid]

    def get_wid(self, word):
        return self._wids.get(word, 0)

    def globToWordIds(self, pattern):
        # Implement * and ? just as in the shell, except the pattern
        # must not start with either of these
        prefix = ""
        while pattern and pattern[0] not in "*?":
            prefix += pattern[0]
            pattern = pattern[1:]
        if not pattern:
            # There were no globbing characters in the pattern
            wid = self._wids.get(prefix, 0)
            if wid:
                return [wid]
            else:
                return []
        if not prefix:
            # The pattern starts with a globbing character.
            # This is too efficient, so we raise an exception.
            raise QueryError(
                "pattern %r shouldn't start with glob character" % pattern)
        pat = prefix
        for c in pattern:
            if c == "*":
                pat += ".*"
            elif c == "?":
                pat += "."
            else:
                pat += re.escape(c)
        pat += "$"
        prog = re.compile(pat)
        keys = self._wids.keys(prefix)  # Keys starting at prefix
        wids = []
        for key in keys:
            if not key.startswith(prefix):
                break
            if prog.match(key):
                wids.append(self._wids[key])
        return wids

    def _getWordIdCreate(self, word):
        wid = self._wids.get(word)
        if wid is None:
            # WidCode requires us to use at least 0x4000 as a base number.
            # The algorithm in versions before 2.13 used the length as a base
            # number. So we don't even try to generate numbers below the
            # length as they are likely all taken
            minimum = 0x4000
            if self._wid_length_based:
                minimum = max(self.length(), 0x4000)

            while True:
                if self._v_nextid is None:
                    self._v_nextid = randrange(minimum, 0x10000000)

                wid = self._v_nextid
                self._v_nextid += 1

                if wid not in self._words:
                    break

                self._v_nextid = None

            self.length.change(1)
            self._wids[word] = wid
            self._words[wid] = word
        return wid


def _text2list(text):
    # Helper: splitter input may be a string or a list of strings
    try:
        text + ""
    except Exception:
        return text
    else:
        return [text]

# Sample pipeline elements


class Splitter(object):

    import re
    if six.PY2:
        rx = re.compile(br"(?L)\w+")
        rxGlob = re.compile(br"(?L)\w+[\w*?]*")
    else:
        rx = re.compile(r"\w+")
        rxGlob = re.compile(r"\w+[\w*?]*")  # See globToWordIds() above

    def process(self, lst):
        result = []
        for s in lst:
            result += self.rx.findall(s)
        return result

    def processGlob(self, lst):
        result = []
        for s in lst:
            result += self.rxGlob.findall(s)
        return result


element_factory.registerFactory('Word Splitter',
                                'Whitespace splitter',
                                Splitter)


class CaseNormalizer(object):

    def process(self, lst):
        return [w.lower() for w in lst]


element_factory.registerFactory('Case Normalizer',
                                'Case Normalizer',
                                CaseNormalizer)

element_factory.registerFactory('Stop Words',
                                ' Don\'t remove stop words',
                                None)


class StopWordRemover(object):

    dict = get_stopdict().copy()

    def process(self, lst):
        return [w for w in lst if w not in self.dict]


element_factory.registerFactory('Stop Words',
                                'Remove listed stop words only',
                                StopWordRemover)


class StopWordAndSingleCharRemover(StopWordRemover):

    dict = get_stopdict().copy()
    for c in range(255):
        dict[chr(c)] = None


element_factory.registerFactory('Stop Words',
                                'Remove listed and single char words',
                                StopWordAndSingleCharRemover)



##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################


def process(stopdict, lst):
    """
    Remove stop words (the keys of dict) from the input list of strings
    to create a new list.

    BBB ZCTextIndex 4.0: Compatibility for former C implementation.
    """
    if not isinstance(stopdict, dict):
        raise TypeError('process requires a dict as argument 1')

    return [w for w in lst if w not in stopdict]



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from BTrees.IIBTree import IISet


class Dummy(object):

    def __init__(self, docid, truth):
        self.id = docid
        self.truth = truth


class TestBooleanIndex(unittest.TestCase):

    def _getTargetClass(self):
        from Products.PluginIndexes.BooleanIndex import BooleanIndex
        return BooleanIndex.BooleanIndex

    def _makeOne(self, attr='truth'):
        return self._getTargetClass()(attr)

    def test_index_true(self):
        index = self._makeOne()
        obj = Dummy(1, True)
        index._index_object(obj.id, obj, attr='truth')
        self.assertTrue(1 in index._unindex)
        self.assertFalse(1 in index._index)

    def test_index_false(self):
        index = self._makeOne()
        obj = Dummy(1, False)
        index._index_object(obj.id, obj, attr='truth')
        self.assertTrue(1 in index._unindex)
        self.assertFalse(1 in index._index)

    def test_index_missing_attribute(self):
        index = self._makeOne()
        obj = Dummy(1, True)
        index._index_object(obj.id, obj, attr='missing')
        self.assertFalse(1 in index._unindex)
        self.assertFalse(1 in index._index)

    def test_search_true(self):
        index = self._makeOne()
        obj = Dummy(1, True)
        index._index_object(obj.id, obj, attr='truth')
        obj = Dummy(2, False)
        index._index_object(obj.id, obj, attr='truth')

        res, idx = index._apply_index({'truth': True})
        self.assertEqual(idx, ('truth', ))
        self.assertEqual(list(res), [1])

    def test_search_false(self):
        index = self._makeOne()
        obj = Dummy(1, True)
        index._index_object(obj.id, obj, attr='truth')
        obj = Dummy(2, False)
        index._index_object(obj.id, obj, attr='truth')

        res, idx = index._apply_index({'truth': False})
        self.assertEqual(idx, ('truth', ))
        self.assertEqual(list(res), [2])

    def test_search_inputresult(self):
        index = self._makeOne()
        obj = Dummy(1, True)
        index._index_object(obj.id, obj, attr='truth')
        obj = Dummy(2, False)
        index._index_object(obj.id, obj, attr='truth')
        obj = Dummy(3, True)
        index._index_object(obj.id, obj, attr='truth')

        # The less common value is indexed
        self.assertEqual(index._index_value, 0)

        res, idx = index._apply_index({'truth': True}, resultset=IISet([]))
        self.assertEqual(idx, ('truth', ))
        self.assertEqual(list(res), [])

        res, idx = index._apply_index({'truth': False}, resultset=IISet([]))
        self.assertEqual(idx, ('truth', ))
        self.assertEqual(list(res), [])

        res, idx = index._apply_index({'truth': True}, resultset=IISet([1]))
        self.assertEqual(list(res), [1])

        res, idx = index._apply_index({'truth': False}, resultset=IISet([1]))
        self.assertEqual(list(res), [])

        res, idx = index._apply_index({'truth': True}, resultset=IISet([2]))
        self.assertEqual(list(res), [])

        res, idx = index._apply_index({'truth': False}, resultset=IISet([2]))
        self.assertEqual(list(res), [2])

        res, idx = index._apply_index({'truth': True},
                                      resultset=IISet([1, 2]))
        self.assertEqual(list(res), [1])

        res, idx = index._apply_index({'truth': False},
                                      resultset=IISet([1, 2]))
        self.assertEqual(list(res), [2])

        res, idx = index._apply_index({'truth': True},
                                      resultset=IISet([1, 3]))
        self.assertEqual(list(res), [1, 3])

        res, idx = index._apply_index({'truth': False},
                                      resultset=IISet([1, 3]))
        self.assertEqual(list(res), [])

        res, idx = index._apply_index({'truth': True},
                                      resultset=IISet([1, 2, 3]))
        self.assertEqual(list(res), [1, 3])

        res, idx = index._apply_index({'truth': False},
                                      resultset=IISet([1, 2, 3]))
        self.assertEqual(list(res), [2])

        res, idx = index._apply_index({'truth': True},
                                      resultset=IISet([1, 2, 99]))
        self.assertEqual(list(res), [1])

        res, idx = index._apply_index({'truth': False},
                                      resultset=IISet([1, 2, 99]))
        self.assertEqual(list(res), [2])

    def test_index_many_true(self):
        index = self._makeOne()
        for i in range(0, 100):
            obj = Dummy(i, i < 80 and True or False)
            index._index_object(obj.id, obj, attr='truth')
        self.assertEqual(list(index._index), list(range(80, 100)))
        self.assertEqual(len(index._unindex), 100)

        res, idx = index._apply_index({'truth': True})
        self.assertEqual(list(res), list(range(0, 80)))
        res, idx = index._apply_index({'truth': False})
        self.assertEqual(list(res), list(range(80, 100)))

    def test_index_many_false(self):
        index = self._makeOne()
        for i in range(0, 100):
            obj = Dummy(i, i >= 80 and True or False)
            index._index_object(obj.id, obj, attr='truth')
        self.assertEqual(list(index._index), list(range(80, 100)))
        self.assertEqual(len(index._unindex), 100)

        res, idx = index._apply_index({'truth': False})
        self.assertEqual(list(res), list(range(0, 80)))
        res, idx = index._apply_index({'truth': True})
        self.assertEqual(list(res), list(range(80, 100)))

    def test_index_many_change(self):
        index = self._makeOne()

        def add(i, value):
            obj = Dummy(i, value)
            index._index_object(obj.id, obj, attr='truth')

        # First lets index only True values
        for i in range(0, 4):
            add(i, True)
        self.assertEqual(list(index._index), [])
        self.assertEqual(len(index._unindex), 4)
        # Now add an equal number of False values
        for i in range(4, 8):
            add(i, False)
        self.assertEqual(list(index._index), list(range(4, 8)))
        self.assertEqual(len(index._unindex), 8)
        # Once False gets to be more than 60% of the indexed set, we switch
        add(8, False)
        self.assertEqual(list(index._index), list(range(4, 9)))
        add(9, False)
        self.assertEqual(list(index._index), list(range(0, 4)))
        res, idx = index._apply_index({'truth': True})
        self.assertEqual(list(res), list(range(0, 4)))
        res, idx = index._apply_index({'truth': False})
        self.assertEqual(list(res), list(range(4, 10)))
        # and we can again switch if the percentages change again
        for i in range(6, 10):
            index.unindex_object(i)
        self.assertEqual(list(index._index), list(range(4, 6)))
        self.assertEqual(len(index._unindex), 6)
        res, idx = index._apply_index({'truth': True})
        self.assertEqual(list(res), list(range(0, 4)))
        res, idx = index._apply_index({'truth': False})
        self.assertEqual(list(res), list(range(4, 6)))

    def test_items(self):
        index = self._makeOne()
        # test empty
        items = dict(index.items())
        self.assertEqual(len(items[True]), 0)
        self.assertEqual(len(items[False]), 0)
        # test few trues
        for i in range(0, 20):
            obj = Dummy(i, i < 5 and True or False)
            index._index_object(obj.id, obj, attr='truth')
        items = dict(index.items())
        self.assertEqual(len(items[True]), 5)
        self.assertEqual(len(items[False]), 15)
        # test many trues
        for i in range(7, 20):
            index.unindex_object(i)
        items = dict(index.items())
        self.assertEqual(len(items[True]), 5)
        self.assertEqual(len(items[False]), 2)

    def test_histogram(self):
        index = self._makeOne()
        # test empty
        hist = index.histogram()
        self.assertEqual(hist[True], 0)
        self.assertEqual(hist[False], 0)
        # test few trues
        for i in range(0, 20):
            obj = Dummy(i, i < 5 and True or False)
            index._index_object(obj.id, obj, attr='truth')
        hist = index.histogram()
        self.assertEqual(hist[True], 5)
        self.assertEqual(hist[False], 15)

    def test_reindexation_when_index_reversed(self):
        index = self._makeOne()
        obj1 = Dummy(1, False)
        index._index_object(obj1.id, obj1, attr='truth')
        obj2 = Dummy(2, False)
        self.assertTrue(index._index_value)
        index._index_object(obj2.id, obj2, attr='truth')
        obj3 = Dummy(3, True)
        index._index_object(obj3.id, obj3, attr='truth')
        obj4 = Dummy(4, True)
        index._index_object(obj4.id, obj4, attr='truth')
        obj1.truth = True
        index._index_object(obj1.id, obj1, attr='truth')
        self.assertFalse(index._index_value)

        res = index._apply_index({'truth': True})[0]
        self.assertEqual(list(index._index), [2])
        self.assertEqual(list(res), [1, 3, 4])

    def test_getCounter(self):
        index = self._makeOne()

        self.assertEqual(index.getCounter(), 0)

        obj = Dummy(1, True)
        index.index_object(obj.id, obj)
        self.assertEqual(index.getCounter(), 1)

        index.unindex_object(obj.id)
        self.assertEqual(index.getCounter(), 2)

        # unknown id
        index.unindex_object(1234)
        self.assertEqual(index.getCounter(), 2)

        # clear is a change
        index.clear()
        self.assertEqual(index.getCounter(), 3)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
from unittest import TestCase

import transaction
from BTrees.Length import Length
from ZODB.DB import DB
from ZODB.FileStorage import FileStorage

from Products.ZCTextIndex.CosineIndex import CosineIndex
from Products.ZCTextIndex.Lexicon import Lexicon
from Products.ZCTextIndex.Lexicon import Splitter
from Products.ZCTextIndex.OkapiIndex import OkapiIndex


class IndexTest(object):
    # Subclasses must set a class variable IndexFactory to the appropriate
    # index object constructor.
    IndexFactory = None

    def setUp(self):
        self.lexicon = Lexicon(Splitter())
        self.index = self.IndexFactory(self.lexicon)

    def test_index_document(self, docid=1):
        doc = 'simple document contains five words'
        self.assertFalse(self.index.has_doc(docid))
        self.index.index_doc(docid, doc)
        self.assertTrue(self.index.has_doc(docid))
        self.assertTrue(self.index._docweight[docid])
        self.assertEqual(len(self.index._docweight), 1)
        self.assertEqual(
            len(self.index._docweight), self.index.document_count())
        self.assertEqual(len(self.index._wordinfo), 5)
        self.assertEqual(len(self.index._docwords), 1)
        self.assertEqual(len(self.index.get_words(docid)), 5)
        self.assertEqual(len(self.index._wordinfo),
                         self.index.length())
        for map in self.index._wordinfo.values():
            self.assertEqual(len(map), 1)
            self.assertIn(docid, map)

    def test_unindex_document(self):
        docid = 1
        self.test_index_document(docid)
        self.index.unindex_doc(docid)
        self.assertEqual(len(self.index._docweight), 0)
        self.assertEqual(
            len(self.index._docweight), self.index.document_count())
        self.assertEqual(len(self.index._wordinfo), 0)
        self.assertEqual(len(self.index._docwords), 0)
        self.assertEqual(len(self.index._wordinfo),
                         self.index.length())

    def test_index_two_documents(self):
        self.test_index_document()
        doc = 'another document just four'
        docid = 2
        self.index.index_doc(docid, doc)
        self.assertTrue(self.index._docweight[docid])
        self.assertEqual(len(self.index._docweight), 2)
        self.assertEqual(
            len(self.index._docweight), self.index.document_count())
        self.assertEqual(len(self.index._wordinfo), 8)
        self.assertEqual(len(self.index._docwords), 2)
        self.assertEqual(len(self.index.get_words(docid)), 4)
        self.assertEqual(len(self.index._wordinfo),
                         self.index.length())
        wids = self.lexicon.termToWordIds('document')
        self.assertEqual(len(wids), 1)
        document_wid = wids[0]
        for wid, map in self.index._wordinfo.items():
            if wid == document_wid:
                self.assertEqual(len(map), 2)
                self.assertIn(1, map)
                self.assertIn(docid, map)
            else:
                self.assertEqual(len(map), 1)

    def test_index_two_unindex_one(self):
        # index two documents, unindex one, and test the results
        self.test_index_two_documents()
        self.index.unindex_doc(1)
        docid = 2
        self.assertEqual(len(self.index._docweight), 1)
        self.assertEqual(
            len(self.index._docweight), self.index.document_count())
        self.assertTrue(self.index._docweight[docid])
        self.assertEqual(len(self.index._wordinfo), 4)
        self.assertEqual(len(self.index._docwords), 1)
        self.assertEqual(len(self.index.get_words(docid)), 4)
        self.assertEqual(len(self.index._wordinfo),
                         self.index.length())
        for map in self.index._wordinfo.values():
            self.assertEqual(len(map), 1)
            self.assertIn(docid, map)

    def test_index_duplicated_words(self, docid=1):
        doc = 'very simple repeat repeat repeat document test'
        self.index.index_doc(docid, doc)
        self.assertTrue(self.index._docweight[docid])
        self.assertEqual(len(self.index._wordinfo), 5)
        self.assertEqual(len(self.index._docwords), 1)
        self.assertEqual(len(self.index.get_words(docid)), 7)
        self.assertEqual(len(self.index._wordinfo),
                         self.index.length())
        self.assertEqual(
            len(self.index._docweight), self.index.document_count())
        wids = self.lexicon.termToWordIds('repeat')
        self.assertEqual(len(wids), 1)
        for wid, map in self.index._wordinfo.items():
            self.assertEqual(len(map), 1)
            self.assertIn(docid, map)

    def test_simple_query_oneresult(self):
        self.index.index_doc(1, 'not the same document')
        results = self.index.search('document')
        self.assertEqual(list(results.keys()), [1])

    def test_simple_query_noresults(self):
        self.index.index_doc(1, 'not the same document')
        results = self.index.search('frobnicate')
        self.assertEqual(list(results.keys()), [])

    def test_query_oneresult(self):
        self.index.index_doc(1, 'not the same document')
        self.index.index_doc(2, 'something about something else')
        results = self.index.search('document')
        self.assertEqual(list(results.keys()), [1])

    def test_search_phrase(self):
        self.index.index_doc(1, 'the quick brown fox jumps over the lazy dog')
        self.index.index_doc(2, 'the quick fox jumps lazy over the brown dog')
        results = self.index.search_phrase('quick brown fox')
        self.assertEqual(list(results.keys()), [1])

    def test_search_glob(self):
        self.index.index_doc(1, 'how now brown cow')
        self.index.index_doc(2, 'hough nough browne cough')
        self.index.index_doc(3, 'bar brawl')
        results = self.index.search_glob('bro*')
        self.assertEqual(list(results.keys()), [1, 2])
        results = self.index.search_glob('b*')
        self.assertEqual(list(results.keys()), [1, 2, 3])


class CosineIndexTest(IndexTest, TestCase):
    IndexFactory = CosineIndex


class OkapiIndexTest(IndexTest, TestCase):
    IndexFactory = OkapiIndex


class TestIndexConflict(TestCase):

    db = None

    def tearDown(self):
        if self.db is not None:
            self.db.close()
            self.storage.cleanup()

    def openDB(self):
        n = 'fs_tmp__{0}'.format(os.getpid())
        self.storage = FileStorage(n)
        self.db = DB(self.storage)

    def test_index_doc_conflict(self):
        self.index = OkapiIndex(Lexicon())
        self.openDB()
        r1 = self.db.open().root()
        r1['i'] = self.index
        transaction.commit()

        r2 = self.db.open().root()
        copy = r2['i']
        # Make sure the data is loaded
        list(copy._docweight.items())
        list(copy._docwords.items())
        list(copy._wordinfo.items())
        list(copy._lexicon._wids.items())
        list(copy._lexicon._words.items())

        self.assertEqual(self.index._p_serial, copy._p_serial)

        self.index.index_doc(0, 'The time has come')
        transaction.commit()

        copy.index_doc(1, 'That time has gone')
        transaction.commit()

    def test_reindex_doc_conflict(self):
        self.index = OkapiIndex(Lexicon())
        self.index.index_doc(0, 'Sometimes change is good')
        self.index.index_doc(1, 'Then again, who asked')
        self.openDB()
        r1 = self.db.open().root()
        r1['i'] = self.index
        transaction.commit()

        r2 = self.db.open().root()
        copy = r2['i']
        # Make sure the data is loaded
        list(copy._docweight.items())
        list(copy._docwords.items())
        list(copy._wordinfo.items())
        list(copy._lexicon._wids.items())
        list(copy._lexicon._words.items())

        self.assertEqual(self.index._p_serial, copy._p_serial)

        self.index.index_doc(0, 'Sometimes change isn\'t bad')
        transaction.commit()

        copy.index_doc(1, 'Then again, who asked you?')
        transaction.commit()


class TestUpgrade(TestCase):

    def test_query_before_totaldoclen_upgrade(self):
        self.index1 = OkapiIndex(Lexicon(Splitter()))
        self.index1.index_doc(0, 'The quiet of night')
        # Revert index1 back to a long to simulate an older index instance
        self.index1._totaldoclen = int(self.index1._totaldoclen())
        self.assertEqual(len(self.index1.search('night')), 1)

    def test_upgrade_totaldoclen(self):
        self.index1 = OkapiIndex(Lexicon())
        self.index2 = OkapiIndex(Lexicon())
        self.index1.index_doc(0, 'The quiet of night')
        self.index2.index_doc(0, 'The quiet of night')
        # Revert index1 back to a long to simulate an older index instance
        self.index1._totaldoclen = int(self.index1._totaldoclen())
        self.index1.index_doc(1, 'gazes upon my shadow')
        self.index2.index_doc(1, 'gazes upon my shadow')
        self.assertEqual(
            self.index1._totaldoclen(), self.index2._totaldoclen())
        self.index1._totaldoclen = int(self.index1._totaldoclen())
        self.index1.unindex_doc(0)
        self.index2.unindex_doc(0)
        self.assertEqual(
            self.index1._totaldoclen(), self.index2._totaldoclen())

    def test_query_before_document_count_upgrade(self):
        self.index1 = OkapiIndex(Lexicon(Splitter()))
        self.index1.index_doc(0, 'The quiet of night')
        # Revert index1 back to a long to simulate an older index instance
        del self.index1.document_count
        self.assertEqual(len(self.index1.search('night')), 1)

    def test_upgrade_document_count(self):
        self.index1 = OkapiIndex(Lexicon())
        self.index2 = OkapiIndex(Lexicon())
        self.index1.index_doc(0, 'The quiet of night')
        self.index2.index_doc(0, 'The quiet of night')
        # Revert index1 back to simulate an older index instance
        del self.index1.document_count
        self.index1.index_doc(1, 'gazes upon my shadow')
        self.index2.index_doc(1, 'gazes upon my shadow')
        self.assertIs(self.index1.document_count.__class__, Length)
        self.assertEqual(
            self.index1.document_count(), self.index2.document_count())
        del self.index1.document_count
        self.index1.unindex_doc(0)
        self.index2.unindex_doc(0)
        self.assertIs(self.index1.document_count.__class__, Length)
        self.assertEqual(
            self.index1.document_count(), self.index2.document_count())
